# -*- coding: utf-8 -*-
import random
import numpy as np
from collections import Counter
import warnings
import ctypes

from qlazy.config import *
from qlazy.error import *
from qlazy.MData import *
from qlazy.Observable import *
from qlazy.PauliProduct import *
from qlazy.lib.qstate_mcx import *

class QState(ctypes.Structure):
    """ Quantum State

    Attributes
    ----------
    qubit_num : int
        qubit number of the quantum state (= log(state_num)).
    state_num : int
        dimension of the quantum state vector (= 2**qubit_num).
    amp : list of complex
        elements of the quantum state vector.

    """

    _fields_ = [
        ('qubit_num', ctypes.c_int),
        ('state_num', ctypes.c_int),
        ('buf_id', ctypes.c_int),
        ('camp', ctypes.c_void_p),
        ('buffer_0', ctypes.c_void_p),
        ('buffer_1', ctypes.c_void_p),
        ('gbank', ctypes.c_void_p),
    ]

    def __new__(cls, qubit_num=None, vector=None, seed=None, **kwargs):
        """
        Parameters
        ----------
        qubit_num : int
            qubit number of the quantum state.
        vector : list
            elements of the quantum state vector.
        seed : int, default - set randomly
            seed for random generation for meaurement.

        Notes
        -----
        You must specify either 'qubit_num' or 'vector', not both.

        """
        if seed is None:
            seed = random.randint(0,1000000)

        if qubit_num is not None:
            if qubit_num > MAX_QUBIT_NUM:
                print("qubit number must be {0:d} or less.".format(MAX_QUBIT_NUM))
                raise QState_Error_Initialize()

            obj = qstate_init(qubit_num, seed)

        else:
            obj = qstate_init_with_vector(vector, seed)
            
        self = ctypes.cast(obj.value, ctypes.POINTER(cls)).contents
        return self

    def __str__(self):

        return str(self.get_amp())

    def reset(self, qid=[]):
        """
        reset to |00..0> state.

        Parameters
        ----------
        qid : list, default - qubit id's list for all of the qubits
            qubit id's list to reset.

        Returns
        -------
        self : instance of QState
            quantum state.

        Notes
        -----
        If 'qid' is set, specified qubits are reset after
        measurement. So if the specified qubits are entangled with the
        remaining qubits, output quantum state is probabilistic. If no
        qubits are set, all qubits are zero reset.

        """
        qstate_reset(self, qid=qid)
        return self
        
    @classmethod
    def add_method(cls, method):
        """
        add method (custum gate).

        Parameters
        ----------
        method : func
            method (custum gate) to add.

        Examples
        --------
        >>> def bell(self, q0, q1):
        >>>     self.h(q0).cx(q0,q1)
        >>>     return self
        >>> ...
        >>> QState.add_method(bell)
        >>> qs = QState(qubit_num=2)
        >>> qs.bell(0,1)
        >>> ...
        
        """
        setattr(cls, method.__name__, method)
        
    @classmethod
    def add_methods(cls, *methods):
        """
        add methods (custum gates).

        Parameters
        ----------
        methods : func, func, ...
            arguments of methods (custum gates) to add.

        Examples
        --------
        >>> def bell(self, q0, q1):
        >>>     self.h(q0).cx(q0,q1)
        >>>     return self
        >>> ...
        >>> def flip(self, q0, q1):
        >>>     self.x(q0).x(q1)
        >>>     return self
        >>> ...
        >>> QState.add_methods(bell, flip, ...)
        >>> qs = QState(2)
        >>> qs.bell(0,1)
        >>> qs.flip(0,1)
        >>> ...
        
        """
        for method in methods:
            if callable(method):
                setattr(cls, method.__name__, method)
            else:
                raise QState_Error_AddMethods()
            
    @classmethod
    def create_register(cls, num):
        """
        create registers (qubit id's) and initialize zero.

        Parameters
        ----------
        num : int
            qubit number you want to use.

        Examples
        --------
        >>> qid = QState.create_register(3)
        >>> print(qid)
        [0,0,0]
        
        """
        qid = [0]*num
        return qid

    @classmethod
    def init_register(cls, *args):
        """
        initialize registers (qubit id's).

        Parameters
        ----------
        args : list, list,...
            arguments of qubit registers.

        Returns
        -------
        idx : int
            total qubit number.

        Examples
        --------
        >>> qid_0 = QState.create_register(3)
        >>> qid_1 = QState.create_register(2)
        >>> print(qid_0, qid_1)
        [0,0,0] [0,0]
        >>> qnum = QState.init_register(qid_0, qid_1)
        >>> print(qnum, qid_0, qid_1)
        5 [0,1,2] [3,4]
        """
        idx = 0
        for i in range(len(args)):
            for j in range(len(args[i])):
                args[i][j] = idx
                idx += 1
        return idx
    
    @classmethod
    def free_all(cls, *qstates):
        """
        free memory of the all quantum states.

        Parameters
        ----------
        qstates : instance of QState,instance of QState,...
            set of QState instances

        Returns
        -------
        None

        """
        warnings.warn("No need to call 'free_all' method because free automatically, or you can use class method 'del_all' to free memory explicitly.")
        
    @classmethod
    def del_all(cls, *qstates):
        """
        free memory of the all quantum states.

        Parameters
        ----------
        qstates : instance of QState,instance of QState,...
            set of QState instances

        Returns
        -------
        None

        """
        for qs in qstates:
            if type(qs) is list or type(qs) is tuple:
                cls.del_all(*qs)
            elif type(qs) is QState:
                del qs
            else:
                raise QState_Error_FreeAll()

    @property
    def amp(self, qid=None):
        """ elements of quantum state vector. """
        return self.get_amp()
        
    def get_amp(self, qid=None):
        """
        get the elements of quantum state vector.

        Parameters
        ----------
        qid : list of int, default - list of all of the qubit id
            qubit id's list.

        Returns
        -------
        ret : list of complex
            elements of the quantum state vector.

        Notes
        -----
        If 'qid' is set, specified qubit state are got after remaining
        qubits are measured. So if the specified qubits are entangled
        with the remaining qubits, output quantum state is
        probabilistic. If no qubits are specified, you get all
        elements of the quantum state.

        """
        ret =  qstate_get_camp(self, qid)
        return ret

    def partial(self, qid=None):
        """
        get the partial quantum state.

        Parameters
        ----------
        qid : list of int, default - list of all of the qubit id
            qubit id's list to get as a partial quantum
            system.

        Returns
        -------
        qs : instance of QState
            partial quantum state.

        Notes
        -----
        If 'qid' is set, specified partial quantum system are got
        after remaining system are measured. So if the specified
        quantum system are entangled with the remaining system, output
        quantum state is probabilistic. If no qubits are specified,
        you get the copy of original quantum state.

        """
        vec = self.get_amp(qid)
        qs = __class__(vector=vec)
        return qs
        
    def show(self, qid=None, nonzero=False):
        """
        show the quantum state 
        (elements of the state vector and probabilities).

        Parameters
        ----------
        qid : list of int, default - list of all of the qubit id
            qubit id's list to show.
        nonzero : bool, default False
            if True, only non-zero amplitudes are printed.

        Returns
        -------
        None

        Notes
        -----
        If 'qid' is set, it shows the elements of quantum state vector
        for partial quantum system specified by 'qid'. If the
        specified quantum system are entangled with the remaining
        system, output is probabilistic. If no qubits are specified,
        it shows the whole quantum state. This method does not change
        the original quantum state.

        Examples
        --------
        >>> qs = QState(2).h(0).cx(0,1)
        >>> qs.show()
        c[00] = +0.7071+0.0000*i : 0.5000 |++++++
        c[01] = +0.0000+0.0000*i : 0.0000 |
        c[10] = +0.0000+0.0000*i : 0.0000 |
        c[11] = +0.7071+0.0000*i : 0.5000 |++++++
        ...
        >>> qs.show(qid=[0])
        c[0] = +1.0000+0.0000*i : 1.0000 |+++++++++++
        c[1] = +0.0000+0.0000*i : 0.0000 |
        ...
        >>> qs.show(nonzero=True)
        c[00] = +0.7071+0.0000*i : 0.5000 |++++++
        c[11] = +0.7071+0.0000*i : 0.5000 |++++++

        """
        qstate_print(self, qid, nonzero)

    def clone(self):
        """
        get the copy of the quantum state.

        Parameters
        ----------
        None

        Returns
        -------
        qstate : instance of QState
            copy of the original quantum state.

        """
        obj = qstate_copy(self)
        qs = ctypes.cast(obj.value, ctypes.POINTER(self.__class__)).contents
        return qs

    def bloch(self, q=0):
        """
        get bloch angles.

        Parameters
        ----------
        q : int
            qubit id

        Returns
        -------
        theta : float
            bloch angle with Z-axis
        phi : float
            bloch angle with X-axis

        Notes
        -----
        the unit of angle is PI radian. for example, 0.5 means 0.5*PI
        (radian).

        """
        theta, phi = qstate_bloch(self, q=0)
        return theta, phi

    def inpro(self, qstate, qid=[]):
        """
        get the inner product with quantum state.

        Parameters
        ----------
        qstate : instance of QState
            one of the two quantum state.
        qid : list of int, default - list of all of the qubit id
            qubit id's list.

        Returns
        -------
        inp : complex
            inner produt (<self|qstate>).

        Notes
        -----
        If 'qid' is set, you can get the inner product for partial
        quantum state. If the specified quantum system are entangled
        with the remaining system, output value is probabilistic,
        while original quantum states do not change.

        """
        if qid == []:
            inp = qstate_inner_product(self, qstate)
        else:
            qs_0 = self.partial(qid=qid)
            qs_1 = qstate.partial(qid=qid)
            inp = qstate_inner_product(qs_0, qs_1)
        return inp
        
    def tenspro(self, qstate):
        """
        get the tensor product with quantum state.

        Parameters
        ----------
        qstate : instance of QState
            quantum state to get the tensor product.

        Returns
        -------
        qstate_out : instance of QState
            tensor produt of 'self' and 'qstate'.

        """
        obj = qstate_tensor_product(self, qstate)
        qs = ctypes.cast(obj.value, ctypes.POINTER(self.__class__)).contents
        return qs

    def fidelity(self, qstate, qid=[]):
        """
        get the fidelity with quantum state.

        Parameters
        ----------
        qstate : instance of QState
            one of the two quantum state.
        qid : list of int
            qubit id's list.

        Returns
        -------
        fid : float
            fidelity of two quantum states. absolute value of the
            inner product of two quantum states.

        Notes
        -----
        If 'qid' is set, you can get the fidelity for partial quantum
        state. If the specified quantum system are entangled with the
        remaining system, output value is probabilistic, while
        original quantum states do not change.

        """
        return abs(self.inpro(qstate, qid=qid))

    def composite(self, num=1):
        """
        get the composite state of same quantum states.

        Parameters
        ----------
        num : int
            number of quantum states.

        Returns
        -------
        qs : instance of QState
            composite quantum state.

        """
        if num <= 1:
            return self
        else:
            qs = self.clone()
            for i in range(num-1):
                qs_tmp = qs.tenspro(self)
                qs = qs_tmp.clone()
            return qs
        
    def join(self, qs_list):
        """
        get tensor product state of the quantum states' list.

        Parameters
        ----------
        qs_list : list (QState)
            list of quantum states.

        Returns
        -------
        qs_out : instance of QState
            tensor product state.

        """
        qs_out = self.clone()
        for qs in qs_list:
            qs_tmp = qs_out.clone()
            qs_out = qs_tmp.tenspro(qs)
        return qs_out

    def evolve(self, observable=None, time=0.0, iter=0):
        """
        evolve the quantum state.

        Parameters
        ----------
        observable : instance of Observable
            Hamiltonian of the system.
        time : float
            period of time.
        iter : int
            number of iteration.

        Returns
        -------
        self : instance of QState

        Notes
        -----
        The 'iter' value should be sufficiently larger than the
        'time' value. This method change the original state.

        See Also
        --------
        Obserbable class (Observable.py)

        """
        qstate_evolve(self, observable=observable, time=time, iter=iter)
        return self
    
    def expect(self, observable=None):
        """
        get the expectation value for observable under the quantum state.

        Parameters
        ----------
        observable : instance of Observable
            obserbable of the system.

        Returns
        -------
        expect : float
            expect value.

        See Also
        --------
        Obserbable class (Observable.py)

        """
        expect = qstate_expect_value(self, observable=observable)
        return expect
    
    def apply(self, matrix=None, qid=None):
        """
        apply matrix.

        Parameters
        ----------
        matrix : list of list
            matrix to apply.

        Returns
        -------
        self : instance of QState

        Notes
        -----
        If 'qid' isn't set, dimension of the matrix must be equal to
        the 2 power of qubit number of the system. If 'qid' is set,
        dimension of the matrix must be equal to the 2 power of 'qid'
        length.

        """
        qstate_apply_matrix(self, matrix=matrix, qid=qid)
        return self

    def __schmidt_decomp(self, qid_0=[], qid_1=[]):

        vec = self.get_amp(qid=qid_0+qid_1)

        row = 2**len(qid_0)
        col = 2**len(qid_1)
        mat = np.zeros((row, col), dtype=complex)
        for idx,comp in enumerate(vec):
            mat[idx//col][idx%col] = comp
    
        U,D,V = np.linalg.svd(mat, full_matrices=False)
        coef = np.array([d for d in D if d > EPS])
        
        vec_0 = [v for i,v in enumerate(U.T) if i < len(coef)]
        vec_1 = [v for i,v in enumerate(V) if i < len(coef)]
    
        return (coef, vec_0, vec_1)

    def schmidt_decomp(self, qid_0=[], qid_1=[]):
        """
        schmidt decomposition.

        Parameters
        ----------
        qid_0 : list
            subsystem to decomposite.
        qid_1 : list
            another subsystem to decomposite.

        Returns
        -------
        coef : numpy.ndarray
            schmidt coefficients.
        qs_0 : list of QState instances
            decomposite quantum state related to 'qid_0'.
        qs_1 : list of QState instances
            decomposite quantum state related to 'qid_1'.

        """
        coef, vec_0, vec_1 = self.__schmidt_decomp(qid_0=qid_0, qid_1=qid_1)
        qs_0 = [__class__(vector=v) for i,v in enumerate(vec_0) if i < len(coef)]
        qs_1 = [__class__(vector=v) for i,v in enumerate(vec_1) if i < len(coef)]
    
        return (coef, qs_0, qs_1)

    def schmidt_coef(self, qid_0=[], qid_1=[]):
        """
        get schmidt coefficients.

        Parameters
        ----------
        qid_0 : list
            subsystem to decomposite.
        qid_1 : list
            another subsystem to decomposite.

        Returns
        -------
        coef : numpy.ndarray
            schmidt coefficients.

        """
        coef, vec_0, vec_1 = self.__schmidt_decomp(qid_0=qid_0, qid_1=qid_1)
        return coef

    # 1-qubit gate

    def x(self, q0):
        """
        operate X gate.

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=PAULI_X, phase=DEF_PHASE, qid=[q0])
        return self

    def y(self, q0):
        """
        operate Y gate.

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=PAULI_Y, phase=DEF_PHASE, qid=[q0])
        return self

    def z(self, q0):
        """
        operate Z gate.

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=PAULI_Z, phase=DEF_PHASE, qid=[q0])
        return self

    def xr(self, q0):
        """
        operate root X gate.

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=ROOT_PAULI_X, phase=DEF_PHASE, qid=[q0])
        return self

    def xr_dg(self, q0):
        """
        operate root X dagger gate 
        (hermmitian conjugate of root X gate).

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=ROOT_PAULI_X_, phase=DEF_PHASE, qid=[q0])
        return self

    def h(self, q0):
        """
        operate H gate (hadamard gate).

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=HADAMARD, phase=DEF_PHASE, qid=[q0])
        return self

    def s(self, q0):
        """
        operate S gate.

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=PHASE_SHIFT_S, phase=DEF_PHASE, qid=[q0])
        return self

    def s_dg(self, q0):
        """
        operate S dagger gate (hermitian conjugate of S gate).

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=PHASE_SHIFT_S_, phase=DEF_PHASE, qid=[q0])
        return self

    def t(self, q0):
        """
        operate T gate.

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=PHASE_SHIFT_T, phase=DEF_PHASE, qid=[q0])
        return self

    def t_dg(self, q0):
        """
        operate T dagger gate (hermitian conjugate of T gate).

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=PHASE_SHIFT_T_, phase=DEF_PHASE, qid=[q0])
        return self

    def rx(self, q0, phase=DEF_PHASE):
        """
        operate RX gate (rotation around X-axis).

        Parameters
        ----------
        q0 : int
            qubit id.
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=ROTATION_X, phase=phase, qid=[q0])
        return self

    def ry(self, q0, phase=DEF_PHASE):
        """
        operate RY gate (rotation around Y-axis).

        Parameters
        ----------
        q0 : int
            qubit id.
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=ROTATION_Y, phase=phase, qid=[q0])
        return self

    def rz(self, q0, phase=DEF_PHASE):
        """
        operate RZ gate (rotation around Z-axis).

        Parameters
        ----------
        q0 : int
            qubit id.
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=ROTATION_Z, phase=phase, qid=[q0])
        return self

    def p(self, q0, phase=DEF_PHASE):
        """
        operate P gate (phase shift gate).

        Parameters
        ----------
        q0 : int
            qubit id.
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of QState

        Notes
        -----
        matrix expression is following...
        | 1.0 0.0             |
        | 0.0 exp(i*phase*PI) |

        """
        qstate_operate_qgate(self, kind=PHASE_SHIFT, phase=phase, qid=[q0])
        return self

    # def u1(self, q0, alpha=DEF_PHASE):
    #     """
    #     operate U1 gate (by IBM).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id.
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    # 
    #     Returns
    #     -------
    #     self : instance of QState
    # 
    #     Notes
    #     -----
    #     this opration is equal to P gate (phase shift gate)
    # 
    #     """
    #     qstate_operate_qgate(self, kind=ROTATION_U1, phase=alpha, qid=[q0])
    #     return self
    # 
    # def u2(self, q0, alpha=DEF_PHASE, beta=DEF_PHASE):
    #     """
    #     operate U2 gate (by IBM).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id.
    #     alpha : float
    #         rotation angle (unit of angle is pi radian).
    #     beta : float
    #         rotation angle (unit of angle is pi radian).
    # 
    #     Returns
    #     -------
    #     self : instance of QState
    # 
    #     Notes
    #     -----
    #     matrix experssion is following...
    #     | 1/sqrt(2)              -exp(i*alpha*PI)/sqrt(2)       |
    #     | exp(i*beta*PI)/sqrt(2) exp(i*(alpha+beta)*PI)/sqrt(2) |
    # 
    #     """
    #     qstate_operate_qgate(self, kind=ROTATION_U2, phase=alpha, phase1=beta, qid=[q0])
    #     return self
    # 
    # def u3(self, q0, alpha=DEF_PHASE, beta=DEF_PHASE, gamma=DEF_PHASE):
    #     """
    #     operate U3 gate (by IBM).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id.
    #     alpha : float
    #         rotation angle (unit of angle is pi radian).
    #     beta : float
    #         rotation angle (unit of angle is pi radian).
    #     gamma : float
    #         rotation angle (unit of angle is pi radian).
    # 
    #     Returns
    #     -------
    #     self : instance of QState
    # 
    #     Notes
    #     -----
    #     matrix expression is following...
    #     | cos(gamma/2)                -exp(i*alpha*PI)*sin(gamma/2)       |
    #     | exp(i*beta*PI)*sin(gamma/2) exp(i*(alpha+beta)*PI)*cos(gamma/2) |
    # 
    # 
    #     """
    #     qstate_operate_qgate(self, kind=ROTATION_U3, phase=alpha, phase1=beta,
    #                          phase2=gamma, qid=[q0])
    #     return self

    # 2-qubit gate

    def cx(self, q0, q1):
        """
        operate CX gate (controlled X gate, controlled NOT gate, CNOT gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_X, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cy(self, q0, q1):
        """
        operate CY gate (controlled X gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_Y, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cz(self, q0, q1):
        """
        operate CZ gate (controlled Z gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_Z, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cxr(self, q0, q1):
        """
        operate CXR gate (controlled root X gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_XR, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cxr_dg(self, q0, q1):
        """
        operate CXR dagger gate (controlled XR dagger gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_XR_, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def ch(self, q0, q1):
        """
        operate CH gate (controlled H gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_H, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cs(self, q0, q1):
        """
        operate CS gate (controlled S gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_S, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cs_dg(self, q0, q1):
        """
        operate CS dagger gate (controlled S dagger gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_S_, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def ct(self, q0, q1):
        """
        operate CT gate (controlled T gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_T, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def ct_dg(self, q0, q1):
        """
        operate CT dagger gate (controlled T dagger gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_T_, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def sw(self, q0, q1):
        """
        swap gate

        Parameters
        ----------
        q0 : int
            qubit id
        q1 : int
            qubit id

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=SWAP_QUBITS, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cp(self, q0, q1, phase=DEF_PHASE):
        """
        operate CP gate (controlled P gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_P, phase=phase, qid=[q0,q1])
        return self

    def crx(self, q0, q1, phase=DEF_PHASE):
        """
        operate CRX gate (controlled RX gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_RX, phase=phase, qid=[q0,q1])
        return self

    def cry(self, q0, q1, phase=DEF_PHASE):
        """
        operate CRY gate (controlled RY gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_RY, phase=phase, qid=[q0,q1])
        return self

    def crz(self, q0, q1, phase=DEF_PHASE):
        """
        operate CRZ gate (controlled RZ gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of QState

        """
        qstate_operate_qgate(self, kind=CONTROLLED_RZ, phase=phase, qid=[q0,q1])
        return self

    # def cu1(self, q0, q1, alpha=DEF_PHASE):
    #     """
    #     operate CU1 gate (controlled U1 gate).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id (control qubit).
    #     q1 : int
    #         qubit id (target qubit).
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    # 
    #     Returns
    #     -------
    #     self : instance of QState
    # 
    #     """
    #     qstate_operate_qgate(self, kind=CONTROLLED_U1, phase=alpha, qid=[q0,q1])
    #     return self
    # 
    # def cu2(self, q0, q1, alpha=DEF_PHASE, beta=DEF_PHASE):
    #     """
    #     operate CU2 gate (controlled U2 gate).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id (control qubit).
    #     q1 : int
    #         qubit id (target qubit).
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    #     beta : float
    #         rotation angle (unit of angle is PI radian).
    # 
    #     Returns
    #     -------
    #     self : instance of QState
    # 
    #     """
    #     qstate_operate_qgate(self, kind=CONTROLLED_U2, phase=alpha, phase1=beta,
    #                          qid=[q0,q1])
    #     return self
    # 
    # def cu3(self, q0, q1, alpha=DEF_PHASE, beta=DEF_PHASE, gamma=DEF_PHASE):
    #     """
    #     operate CU3 gate (controlled U3 gate).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id (control qubit).
    #     q1 : int
    #         qubit id (target qubit).
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    #     beta : float
    #         rotation angle (unit of angle is PI radian).
    #     gamma : float
    #         rotation angle (unit of angle is PI radian).
    # 
    #     Returns
    #     -------
    #     self : instance of QState
    # 
    #     """
    #     qstate_operate_qgate(self, kind=CONTROLLED_U3, phase=alpha, phase1=beta,
    #                          phase2=gamma, qid=[q0,q1])
    #     return self

    # 3-qubit gate
    
    def ccx(self, q0, q1, q2):
        """
        operate CCX gate (toffoli gate, controlled controlled X gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (control qubit).
        q2 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of QState

        """
        self.cxr(q1,q2).cx(q0,q1).cxr_dg(q1,q2).cx(q0,q1).cxr(q0,q2)
        return self

    def csw(self, q0, q1, q2):
        """
        operate CSW gate (fredkin gate, controlled swap gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (swap qubit).
        q2 : int
            qubit id (swap qubit).

        Returns
        -------
        self : instance of QState

        """
        self.cx(q2,q1).ccx(q0,q1,q2).cx(q2,q1)
        return self

    # other gate

    def mcx(self,qid=[]):
        """
        operate MCX gate (multi-controlled X gate).
    
        Parameters
        ----------
        qid : list of int
            qubit id list [control, control, ... , control, target]
    
        Returns
        -------
        self : instance of QState
    
        """
        qstate_mcx(self, qid)
        return self
    
    # measurement

    def measure(self, qid=None):
        """
        one shot measurement in Z-direction.

        Parameters
        ----------
        qid : list of int
            qubit id list to measure.

        Returns
        -------
        mval : str
            measurement value.

        Examples
        --------
        >>> qs = QState(qubit_num=2).h(0).cx(0,1)
        >>> qs.show()
        >>> print(qs.measure(qid=[0,1]))
        >>> qs.show()
        c[00] = +0.7071+0.0000*i : 0.5000 |++++++
        c[01] = +0.0000+0.0000*i : 0.0000 |
        c[10] = +0.0000+0.0000*i : 0.0000 |
        c[11] = +0.7071+0.0000*i : 0.5000 |++++++
        00
        c[00] = +1.0000+0.0000*i : 1.0000 |+++++++++++
        c[01] = +0.0000+0.0000*i : 0.0000 |
        c[10] = +0.0000+0.0000*i : 0.0000 |
        c[11] = +0.0000+0.0000*i : 0.0000 |

        """
        mval = qstate_measure(self, qid=qid, angle=0.0, phase=0.0)
        return mval
    
    def m(self, qid=None, shots=DEF_SHOTS, angle=0.0, phase=0.0):
        """
        measurement in any direction (default: Z-axis).

        Parameters
        ----------
        qid : list of int
            qubit id list to measure.
        shots : int, default 1
            number of measurements.
        angle : float, default 0.0
            direction of measurement (angle with Z-axis).
        phase : float, default 0.0
            direction of measurement (phase around Z-axis).

        Returns
        -------
        md : instance of MData
            measurement data.

        Examples
        --------
        >>> qs = QState(2).h(0).cx(0,1)
        >>> md = qs.m()
        >>> md.show(shots=100)
        direction of measurement: z-axis
        frq[00] = 52
        frq[11] = 48
        last state => 11

        See Also
        --------
        MData class (MData.py)

        """
        md = qstate_measure_stats(self, qid=qid, shots=shots,
                                  angle=angle, phase=phase)
        return md
        
    def mx(self, qid=None, shots=DEF_SHOTS):
        """
        X-axis measurement.

        Parameters
        ----------
        qid : list of int
            qubit id list to measure.
        shots : int, default 1
            number of measurements.

        Returns
        -------
        md : instance of MData
            measurement data.

        See Also
        --------
        MData class (MData.py)

        """
        md = qstate_measure_stats(self, qid=qid, shots=shots,
                                  angle=0.5, phase=0.0)
        return md
        
    def my(self, qid=None, shots=DEF_SHOTS):
        """
        Y-axis measurement.

        Parameters
        ----------
        qid : list of int
            qubit id list to measure.
        shots : int, default 1
            number of measurements.

        Returns
        -------
        md : instance of MData
            measurement data.

        See Also
        --------
        MData class (MData.py)

        """
        md =  qstate_measure_stats(self, qid=qid, shots=shots,
                                   angle=0.5, phase=0.5)
        return md
        
    def mz(self, qid=None, shots=DEF_SHOTS):
        """
        Z-axis measurement.

        Parameters
        ----------
        qid : list of int
            qubit id list to measure.
        shots : int, default 1
            number of measurements.

        Returns
        -------
        md : instance of MData
            measurement data.

        See Also
        --------
        MData class (MData.py)

        """
        md =  qstate_measure_stats(self, qid=qid, shots=shots,
                                   angle=0.0, phase=0.0)
        return md

    def mb(self, qid=None, shots=DEF_SHOTS):
        return qstate_measure_bell_stats(self, qid=qid, shots=shots)
        
    def operate(self, pp=None, ctrl=None):
        """
        operate unitary operator to quantum state.

        Parameters
        ----------
        pp : instance of PauliProduct
            pauli product to operate
        ctrl : int
            contoroll qubit id for controlled pauli product

        Returns
        -------
        self : instance of QState
            quantum state after operation

        """
        pauli_list = pp.pauli_list
        qid = pp.qid
    
        if ctrl is None:
            for q, pauli in zip(qid, pauli_list):
                if pauli == 'X':
                    self.x(q)
                elif pauli == 'Y':
                    self.y(q)
                elif pauli == 'Z':
                    self.z(q)
                else:
                    continue
        else:
            if ctrl in qid:
                raise ValueError("controll and target qubit id conflict")
        
            for q, pauli in zip(qid, pauli_list):
                if pauli == 'X':
                    self.cx(ctrl, q)
                elif pauli == 'Y':
                    self.cy(ctrl, q)
                elif pauli == 'Z':
                    self.cz(ctrl, q)
                else:
                    continue
    
        return self

    def free(self):
        """
        free memory of quantum state.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        warnings.warn("No need to call 'free' method because free automatically, or you can use 'del' to free memory explicitly.")

    def __del__(self):

        qstate_free(self)

# c-library for qstate
from qlazy.lib.qstate_c import *



# -*- coding: utf-8 -*-
import ctypes
from collections import Counter
from ctypes.util import find_library
import warnings
import ctypes

from qlazy.config import *
from qlazy.error import *

class CMem(ctypes.Structure):
    """ Classical Memory

    Attributes
    ----------
    cmem_num : int
        number of the classical register (classical memory size).
    bit_array : list (int)
        bit array of classical memory.

    """
    _fields_ = [
        ('cmem_num', ctypes.c_int),
        ('bit_array', ctypes.c_void_p),
    ]

    def __new__(cls, cmem_num, **kwargs):
        """
        Parameters
        ----------
        cmem_num : int
            number of the classical register.

        Returns
        -------
        cmem : instance (CMem)

        """
        obj = cmem_init(cmem_num)
        cmem = ctypes.cast(obj.value, ctypes.POINTER(cls)).contents
        return cmem
            
    def clone(self):
        """
        get the copy of the classical memory.

        Parameters
        ----------
        None

        Returns
        -------
        cmem : instance of CMem
            copy of the original classical memory.

        """
        obj = cmem_copy(self)
        cmem = ctypes.cast(obj.value, ctypes.POINTER(self.__class__)).contents
        return cmem

    def __del__(self):
        
        cmem_free(self)
        
# c-library for qstate
from qlazy.lib.cmem_c import *



# -*- coding: utf-8 -*-
import ctypes
from ctypes.util import find_library
import math
import numpy as np
import pathlib

from qlazy.config import *
from qlazy.error import *
from qlazy.util import *
from qlazy.QState import QState
from qlazy.DensOp import DensOp

lib= ctypes.CDLL(str(pathlib.Path(__file__).with_name('libqlz.'+get_lib_ext())))
libc = ctypes.CDLL(find_library("c"),mode=ctypes.RTLD_GLOBAL)

def densop_init(qstate=[], prob=[]):
        
    num = len(qstate)

    densop = None
    c_densop = ctypes.c_void_p(densop)

    DoubleArray = ctypes.c_double * num
    prob_array = DoubleArray(*prob)

    QStateArray = QState * num
    qstate_array = QStateArray(*qstate)

    lib.densop_init.restype = ctypes.c_int
    lib.densop_init.argtypes = [ctypes.POINTER(QState),
                                ctypes.POINTER(ctypes.c_double),
                                ctypes.c_int,
                                ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.densop_init(qstate_array, prob_array,
                          ctypes.c_int(num), c_densop)

    if ret == FALSE:
        raise DensOp_Error_Initialize()
            
    return c_densop

def densop_init_with_matrix(matrix=None):

    densop = None
    c_densop = ctypes.c_void_p(densop)

    row = len(matrix)
    col = row
    size = row * col

    mat_complex = list(matrix.flatten())
    mat_real = [0.0 for _ in range(size)]
    mat_imag = [0.0 for _ in range(size)]
    for i in range(size):
        mat_real[i] = mat_complex[i].real
        mat_imag[i] = mat_complex[i].imag
                
    DoubleArray = ctypes.c_double * size
    c_mat_real = DoubleArray(*mat_real)
    c_mat_imag = DoubleArray(*mat_imag)
            
    lib.densop_init_with_matrix.restype = ctypes.c_int
    lib.densop_init_with_matrix.argtypes = [DoubleArray, DoubleArray,
                                            ctypes.c_int, ctypes.c_int,
                                            ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.densop_init_with_matrix(c_mat_real, c_mat_imag,
                                      ctypes.c_int(row), ctypes.c_int(col),
                                      c_densop)
        
    if ret == FALSE:
        raise DensOp_Error_Initialize()
            
    return c_densop


def densop_get_elm(de):

    try:
        elm = None
        c_elm = ctypes.c_void_p(elm)

        lib.densop_get_elm.restype = ctypes.c_int
        lib.densop_get_elm.argtypes = [ctypes.POINTER(DensOp),
                                           ctypes.POINTER(ctypes.c_void_p)]
        ret = lib.densop_get_elm(ctypes.byref(de), c_elm)

        if ret == FALSE:
            raise DensOp_Error_GetElm()
            
        o = ctypes.cast(c_elm.value, ctypes.POINTER(ctypes.c_double))

        size = de.row * de.col
        out = [0] * size
        for i in range(size):
            out[i] = complex(round(o[2*i],8),round(o[2*i+1],8))

        libc.free.argtypes = [ctypes.POINTER(ctypes.c_double)]
        libc.free(o)

        return np.array(out).reshape([de.row, de.col])

    except Exception:
        raise DensOp_Error_GetElm()

def densop_reset(de, qid=None):

    if qid is None or qid == []:
        qnum = int(math.log2(de.row))
        qid = [i for i in range(qnum)]

    try:
        qubit_num = len(qid)
        qubit_id = [0 for _ in range(qubit_num)]
        for i in range(len(qid)):
            qubit_id[i] = qid[i]

        IntArray = ctypes.c_int * qubit_num
        qid_array = IntArray(*qubit_id)
            
        lib.densop_reset.restype = ctypes.c_int
        lib.densop_reset.argtypes = [ctypes.POINTER(DensOp),ctypes.c_int, IntArray]
        ret = lib.densop_reset(ctypes.byref(de),ctypes.c_int(qubit_num), qid_array)

        if ret == FALSE:
            raise DensOp_Error_Reset()

    except Exception:
        raise DensOp_Error_Reset()

def densop_print(de, nonzero=False):

    try:
        lib.densop_print.restype = ctypes.c_int
        lib.densop_print.argtypes = [ctypes.POINTER(DensOp), ctypes.c_bool]
        ret = lib.densop_print(ctypes.byref(de), ctypes.c_bool(nonzero))

        if ret == FALSE:
            raise DensOp_Error_Show()

    except Exception:
        raise DensOp_Error_Show()

def densop_copy(de):

    try:
        densop = None
        c_densop = ctypes.c_void_p(densop)
            
        lib.densop_copy.restype = ctypes.c_int
        lib.densop_copy.argtypes = [ctypes.POINTER(DensOp),
                                    ctypes.POINTER(ctypes.c_void_p)]
        ret = lib.densop_copy(ctypes.byref(de), c_densop)

        if ret == FALSE:
            raise DensOp_Error_Clone()

        return c_densop
        
    except Exception:
        raise DensOp_Error_Clone()

def densop_add(de, densop=None):

    try:
        lib.densop_add.restype = ctypes.c_int
        lib.densop_add.argtypes = [ctypes.POINTER(DensOp), ctypes.POINTER(DensOp)]
        ret = lib.densop_add(ctypes.byref(de), ctypes.byref(densop))

        if ret == FALSE:
            raise DensOp_Error_Add()
        
    except Exception:
        raise DensOp_Error_Add()

    
def densop_mul(de, factor=0.0):

    try:
        lib.densop_mul.restype = ctypes.c_int
        lib.densop_mul.argtypes = [ctypes.POINTER(DensOp), ctypes.c_double]
        ret = lib.densop_mul(ctypes.byref(de), ctypes.c_double(factor))

        if ret == FALSE:
            raise DensOp_Error_Mul()

    except Exception:
        raise DensOp_Error_Mul()

def densop_trace(de):

    try:
        real = 0.0
        imag = 0.0
        c_real = ctypes.c_double(real)
        c_imag = ctypes.c_double(imag)
            
        lib.densop_trace.restype = ctypes.c_int
        lib.densop_trace.argtypes = [ctypes.POINTER(DensOp),
                                     ctypes.POINTER(ctypes.c_double),
                                     ctypes.POINTER(ctypes.c_double)]
        ret = lib.densop_trace(ctypes.byref(de), ctypes.byref(c_real),
                               ctypes.byref(c_imag))

        if ret == FALSE:
            raise DensOp_Error_Trace()

        real = round(c_real.value, 8)
        imag = round(c_imag.value, 8)

        if abs(imag) > EPS:
            raise DensOp_Error_Trace()
                
        return real
        
    except Exception:
        raise DensOp_Error_Trace()

def densop_sqtrace(de):

    try:
        real = 0.0
        imag = 0.0
        c_real = ctypes.c_double(real)
        c_imag = ctypes.c_double(imag)
            
        lib.densop_sqtrace.restype = ctypes.c_int
        lib.densop_sqtrace.argtypes = [ctypes.POINTER(DensOp),
                                       ctypes.POINTER(ctypes.c_double),
                                       ctypes.POINTER(ctypes.c_double)]
        ret = lib.densop_sqtrace(ctypes.byref(de), ctypes.byref(c_real),
                                 ctypes.byref(c_imag))

        if ret == FALSE:
            raise DensOp_Error_SqTrace()

        real = round(c_real.value, 8)
        imag = round(c_imag.value, 8)

        if abs(imag) > EPS:
            raise DensOp_Error_Trace()
                
        return real

    except Exception:
        raise DensOp_Error_SqTrace()

def densop_patrace(de, qid=None):

    try:
        if qid == None:
            raise DensOp_Error_PaTrace()
            
        densop = None
        c_densop = ctypes.c_void_p(densop)
            
        qubit_num = len(qid)
        qubit_id = [0 for _ in range(qubit_num)]
        for i in range(len(qid)):
            qubit_id[i] = qid[i]
        IntArray = ctypes.c_int * qubit_num
        qid_array = IntArray(*qubit_id)

        lib.densop_patrace.restype = ctypes.c_int
        lib.densop_patrace.argtypes = [ctypes.POINTER(DensOp),
                                       ctypes.c_int,IntArray,
                                       ctypes.POINTER(ctypes.c_void_p)]
        ret = lib.densop_patrace(ctypes.byref(de), ctypes.c_int(qubit_num),
                                 qid_array, c_densop)

        if ret == FALSE:
            raise DensOp_Error_PaTrace()
            
        return c_densop

    except Exception:
        raise DensOp_Error_PaTrace()

def densop_tensor_product(de_0, de_1):

    try:
        densop_out = None
        c_densop_out = ctypes.c_void_p(densop_out)

        lib.densop_tensor_product.restype = ctypes.c_int
        lib.densop_tensor_product.argtypes = [ctypes.POINTER(DensOp),
                                              ctypes.POINTER(DensOp),
                                              ctypes.POINTER(ctypes.c_void_p)]
        ret = lib.densop_tensor_product(ctypes.byref(de_0),ctypes.byref(de_1),
                                        c_densop_out)

        if ret == FALSE:
            raise DensOp_Error_TensorProduct()

        return c_densop_out

    except Exception:
        raise DensOp_Error_TensorProduct()
    
def densop_apply_matrix(de, matrix=None, qid=[], dire='both'):

    if matrix is None:
        raise DensOp_Error_Apply()
        
    if qid is None or qid == []:
        qnum = int(math.log2(de.row))
        qid = [i for i in range(qnum)]

    if dire == 'left':
        adire = LEFT
    elif dire == 'right':
        adire = RIGHT
    elif dire == 'both':
        adire = BOTH
    else:
        raise DensOp_Error_Apply()

    try:
        qubit_num = len(qid)
        qubit_id = [0 for _ in range(qubit_num)]
        for i in range(len(qid)):
            qubit_id[i] = qid[i]
        IntArray = ctypes.c_int * qubit_num
        qid_array = IntArray(*qubit_id)

        row = len(matrix) # dimension of the unitary matrix
        col = row
        size = row * col

        # set array of matrix
        mat_complex = []
        for mat_row in matrix:
            for mat_elm in mat_row:
                mat_complex.append(mat_elm)
        
        mat_real = [0.0 for _ in range(size)]
        mat_imag = [0.0 for _ in range(size)]
        for i in range(size):
            mat_real[i] = mat_complex[i].real
            mat_imag[i] = mat_complex[i].imag
                
        DoubleArray = ctypes.c_double * size
        c_mat_real = DoubleArray(*mat_real)
        c_mat_imag = DoubleArray(*mat_imag)
            
        lib.densop_apply_matrix.restype = ctypes.c_int
        lib.densop_apply_matrix.argtypes = [ctypes.POINTER(DensOp),
                                            ctypes.c_int, IntArray,
                                            ctypes.c_int,
                                            DoubleArray, DoubleArray,
                                            ctypes.c_int, ctypes.c_int]
        ret = lib.densop_apply_matrix(ctypes.byref(de),
                                      ctypes.c_int(qubit_num), qid_array,
                                      ctypes.c_int(adire), c_mat_real, c_mat_imag,
                                      ctypes.c_int(row), ctypes.c_int(col))

        if ret == FALSE:
            raise DensOp_Error_Apply()

    except Exception:
        raise DensOp_Error_Apply()

def densop_probability(de, matrix=None, qid=[], matrix_type=None):

    if matrix is None:
        raise DensOp_Error_Probability()
    if (matrix.shape[0] > de.row or matrix.shape[1] > de.col):
        raise DensOp_Error_Probability()
        
    if qid is None or qid == []:
        qnum = int(math.log2(de.row))
        qid = [i for i in range(qnum)]

    if matrix_type == 'kraus':
        mtype = KRAUS
    elif matrix_type == 'povm':
        mtype = POVM
    else:
        raise DensOp_Error_Probability()
            
    try:
        qubit_num = len(qid)
        qubit_id = [0 for _ in range(qubit_num)]
        for i in range(len(qid)):
            qubit_id[i] = qid[i]
        IntArray = ctypes.c_int * qubit_num
        qid_array = IntArray(*qubit_id)

        row = len(matrix) # dimension of the unitary matrix
        col = row
        size = row * col

        # set array of matrix
        mat_complex = list(matrix.flatten())
        mat_real = [0.0 for _ in range(size)]
        mat_imag = [0.0 for _ in range(size)]
        for i in range(size):
            mat_real[i] = mat_complex[i].real
            mat_imag[i] = mat_complex[i].imag
                
        DoubleArray = ctypes.c_double * size
        c_mat_real = DoubleArray(*mat_real)
        c_mat_imag = DoubleArray(*mat_imag)
            
        prob = 0.0
        c_prob = ctypes.c_double(prob)

        lib.densop_probability.restype = ctypes.c_int
        lib.densop_probability.argtypes = [ctypes.POINTER(DensOp),
                                           ctypes.c_int, IntArray,
                                           ctypes.c_int, DoubleArray, DoubleArray,
                                           ctypes.c_int, ctypes.c_int,
                                           ctypes.POINTER(ctypes.c_double)]
        ret = lib.densop_probability(ctypes.byref(de),
                                     ctypes.c_int(qubit_num), qid_array,
                                     ctypes.c_int(mtype), c_mat_real, c_mat_imag,
                                     ctypes.c_int(row), ctypes.c_int(col),
                                     ctypes.byref(c_prob))

        if ret == FALSE:
            raise DensOp_Error_Probability()

        prob = round(c_prob.value, 8)
            
        return prob

    except Exception:
        raise DensOp_Error_Probability()

def densop_operate_qgate(de, kind=None, qid=None,
                         phase=DEF_PHASE, phase1=DEF_PHASE, phase2=DEF_PHASE):

    # error check
    densop_check_args(de, kind=kind, qid=qid, shots=None, angle=None,
                      phase=phase, phase1=phase1, phase2=phase2)

    qubit_id = [0 for _ in range(2)]
    for i in range(len(qid)):
        qubit_id[i] = qid[i]
    IntArray = ctypes.c_int * 2
    qid_array = IntArray(*qubit_id)

    lib.densop_operate_qgate.restype = ctypes.c_int
    lib.densop_operate_qgate.argtypes = [ctypes.POINTER(DensOp), ctypes.c_int,
                                         ctypes.c_double, ctypes.c_double,
                                         ctypes.c_double, IntArray]
    ret = lib.densop_operate_qgate(ctypes.byref(de), ctypes.c_int(kind),
                                   ctypes.c_double(phase), ctypes.c_double(phase1),
                                   ctypes.c_double(phase2), qid_array)

    if ret == FALSE:
        raise DensOp_Error_OperateQGate()

def densop_free(de):

    lib.densop_free.argtypes = [ctypes.POINTER(DensOp)]
    lib.densop_free(ctypes.byref(de))



# -*- coding: utf-8 -*-
import ctypes
import random
from collections import Counter
import warnings
import ctypes

from qlazy.config import *
from qlazy.error import *

class MDataStabilizer:
    """ Measured Data for Stabilizer

    Attributes
    ----------
    frequency : Counter
        frequencies of measured value.
    last : str
        last measured value.
    qid : list of int
        qubit id's list.
    qubit_num : int
        qubit number of the quantum state (= log(state_num)).

    """

    def __init__(self, frequency=None, last=None, qid=None, qubit_num=0):
        
        self.frequency = frequency
        self.last = last
        self.qid = qid
        self.qubit_num = qubit_num
        
class Stabilizer(ctypes.Structure):
    """ Quantum State

    Attributes
    ----------
    qubit_num : int
        number of qubits.
    gene_num : int
        number of generators.
    """

    _fields_ = [
        ('gene_num', ctypes.c_int),
        ('qubit_num', ctypes.c_int),
        ('pauli_factor', ctypes.c_void_p),
        ('check_matrix', ctypes.c_void_p),
    ]

    def __new__(cls, qubit_num=None, gene_num=None, pp_list=None, seed=None, **kwargs):
        """
        Parameters
        ----------
        gene_num : int
            number of generators.
        qubit_num : int
            number of qubit.

        Notes
        -----
        You must specify 'qubit_num'. If 'gene_num' is not specified, 
        'gene_num' is equal to 'qubit_num.'

        """
        if seed is None:
            seed = random.randint(0,1000000)

        if pp_list is not None:
            gene_num = len(pp_list)
            qubit_num = 0
            for pp in pp_list:
                qubit_num = max([qubit_num] + pp.qid)
            qubit_num += 1
            obj = stabilizer_init(gene_num, qubit_num, seed)
            sb = ctypes.cast(obj.value, ctypes.POINTER(cls)).contents
            for i, pp in enumerate(pp_list):
                for j, q in enumerate(pp.qid):
                    sb.set_pauli_op(i, q, pp.pauli_list[j])
            return sb

        else:
            if qubit_num is None:
                raise Stabilizer_Error_Initialize()
            if gene_num is None:
                gene_num = qubit_num
            if qubit_num < 1 or gene_num < 1:
                raise Stabilizer_Error_Initialize()
            obj = stabilizer_init(gene_num, qubit_num, seed)
            sb = ctypes.cast(obj.value, ctypes.POINTER(cls)).contents
            return sb
            
    def __str__(self):

        return self.get_str()

    @classmethod
    def add_method(cls, method):
        """
        add method (custum gate).

        Parameters
        ----------
        method : func
            method (custum gate) to add.
        
        """
        setattr(cls, method.__name__, method)
        
    @classmethod
    def add_methods(cls, *methods):
        """
        add methods (custum gates).

        Parameters
        ----------
        methods : func, func, ...
            arguments of methods (custum gates) to add.
        
        """
        for method in methods:
            if callable(method):
                setattr(cls, method.__name__, method)
            else:
                raise Stabillizer_Error_AddMethods()
            
    def get_str(self):

        str_out = ""
        for i in range(self.gene_num):
            pauli_fac_complex = self.get_pauli_fac(i)

            gene_str = ""
            for j in range(self.qubit_num):
                s = self.get_pauli_op(i,j)
                gene_str += s

            if pauli_fac_complex == 1+0j:
                pauli_fac_str = "  "
            elif pauli_fac_complex == 1j:
                pauli_fac_str = " i"
            elif pauli_fac_complex == -1+0j:
                pauli_fac_str = " -"
            elif pauli_fac_complex == -1j:
                pauli_fac_str = "-i"
            else:
                raise Stabilizer_Error_GetStr()

            str_out += "{0:}{1:}\n".format(pauli_fac_str,gene_str)
                
        return str_out
        
    def show(self):
        """
        show the generators of stabilizer.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

        digits = len(str(self.gene_num - 1))
        s = self.get_str()
        gene_list = s.rstrip().split('\n')
        for i,gene_str in enumerate(gene_list):
            print("g[{0:{digits}d}]:{1:}".format(i,gene_str, digits=digits))

    def reset(self):
        """
        reset to initial generators which are all identity.

        Parameters
        ----------
        None

        Returns
        -------
        self : instance of Stabilizer
            stabilizer state.
 
        """
        for i in range(self.gene_num):
            for j in range(self.qubit_num):
                self.set_pauli_op(i,j,'I')
            self.set_pauli_fac(i,'+1')
        return self
        
    def clone(self):
        """
        get the copy of the quantum state.

        Parameters
        ----------
        None

        Returns
        -------
        stab : instance of Stabilizer
            copy of the original stabilizer.

        """
        obj = stabilizer_copy(self)
        sb = ctypes.cast(obj.value, ctypes.POINTER(self.__class__)).contents
        return sb

    def get_rank(self):

        rank = stabilizer_get_rank(self)
        return rank
    
    def set_all(self, pauli_op_str):
        """
        set all of the qubits same pauli operators.

        Parameters
        ----------
        pauli_op_str : str
            string of pauli operator ('X','Y','Z').

        Returns
        -------
        self : instance of Stabilizer

        Examples
        --------
        >>> sb = Stabilizer(qubit_num=3)
        >>> sb.set_all('Z')
        >>> sb.show()
        g[0]:  ZII
        g[1]:  IZI
        g[2]:  IIZ

        """
        self.reset()
        len = min(self.gene_num, self.qubit_num)
        [self.set_pauli_op(i, i, pauli_op_str) for i in range(len)]
        return self
    
    def set_pauli_fac(self, gene_id, pauli_fac_str):
        """
        set pauli factor of generator ('+1','-1','+i','-i').

        Parameters
        ----------
        gene_id : int
            generator id to set.

        Returns
        -------
        self : instance of Stabilizer

        """
        if pauli_fac_str == "+1" or pauli_fac_str == "1":
            pauli_fac = REAL_PLUS
        elif pauli_fac_str == "+i":
            pauli_fac = IMAG_PLUS
        elif pauli_fac_str == "-1":
            pauli_fac = REAL_MINUS
        elif pauli_fac_str == "-i":
            pauli_fac = IMAG_MINUS
        else:
            raise Stabilizer_Error_SetPauliFac()
            
        stabilizer_set_pauli_fac(self, gene_id, pauli_fac)

        return self

    def get_pauli_fac(self, gene_id):
        """
        get pauli factor of generator ('+1','-1','+i','-i').

        Parameters
        ----------
        gene_id : int
            generator id to get.

        Returns
        -------
        pauli_fac_complex : complex
            complex facto of the generator (1+0j, 1j, -1+0j, -1j)

        """
        pauli_fac = stabilizer_get_pauli_fac(self, gene_id)

        if pauli_fac == REAL_PLUS:
            pauli_fac_complex = 1+0j
        elif pauli_fac == IMAG_PLUS:
            pauli_fac_complex = 1j
        elif pauli_fac == REAL_MINUS:
            pauli_fac_complex = -1+0j
        elif pauli_fac == IMAG_MINUS:
            pauli_fac_complex = -1j
        else:
            raise Stabilizer_Error_GetPauliFac()

        return pauli_fac_complex

    def set_pauli_op(self, gene_id, qubit_id, pauli_op_str):
        """
        set pauli operator ('I','X','Y','Z').

        Parameters
        ----------
        gene_id : int
            generator id to set.
        qubit_id : int
            qubit id to set.

        Returns
        -------
        self : instance of Stabilizer

        """
        if pauli_op_str == 'X':
            pauli_op = PAULI_X
        elif pauli_op_str == 'Y':
            pauli_op = PAULI_Y
        elif pauli_op_str == 'Z':
            pauli_op = PAULI_Z
        elif pauli_op_str == 'I':
            pauli_op = IDENTITY
        else:
            raise Stabilizer_Error_SetPauliOp()
            
        stabilizer_set_pauli_op(self, gene_id, qubit_id, pauli_op)

        return self
    
    def get_pauli_op(self, gene_id, qubit_id):
        """
        get pauli operator ('I','X','Y','Z').

        Parameters
        ----------
        gene_id : int
            generator id to get.
        qubit_id : int
            qubit id to get.

        Returns
        -------
        pauli_op_str : str
            pauli operator ('I','X','Y','Z')

        """
        pauli_op = stabilizer_get_pauli_op(self, gene_id, qubit_id)

        if pauli_op == PAULI_X:
            pauli_op_str = 'X'
        elif pauli_op == PAULI_Y:
            pauli_op_str = 'Y'
        elif pauli_op == PAULI_Z:
            pauli_op_str = 'Z'
        elif pauli_op == IDENTITY:
            pauli_op_str = 'I'
        else:
            raise Stabilizer_Error_GetPauliOp()

        return pauli_op_str

    # 1-qubit gate
    
    def x(self, q):
        """
        operate X gate.

        Parameters
        ----------
        q : int
            qubit id.

        Returns
        -------
        self : instans of Stabilizer

        """
        stabilizer_operate_qgate(self, PAULI_X, q, 0)
        return self
    
    def y(self, q):
        """
        operate Y gate.

        Parameters
        ----------
        q : int
            qubit id.

        Returns
        -------
        self : instans of Stabilizer

        """
        stabilizer_operate_qgate(self, PAULI_Y, q, 0)
        return self
    
    def z(self, q):
        """
        operate Z gate.

        Parameters
        ----------
        q : int
            qubit id.

        Returns
        -------
        self : instans of Stabilizer

        """
        stabilizer_operate_qgate(self, PAULI_Z, q, 0)
        return self
    
    def h(self, q):
        """
        operate H gate.

        Parameters
        ----------
        q : int
            qubit id.

        Returns
        -------
        self : instans of Stabilizer

        """
        stabilizer_operate_qgate(self, HADAMARD, q, 0)
        return self
    
    def s(self, q):
        """
        operate S gate.

        Parameters
        ----------
        q : int
            qubit id.

        Returns
        -------
        self : instans of Stabilizer

        """
        stabilizer_operate_qgate(self, PHASE_SHIFT_S, q, 0)
        return self
    
    def s_dg(self, q):
        """
        operate S dagger gate.

        Parameters
        ----------
        q : int
            qubit id.

        Returns
        -------
        self : instans of Stabilizer

        """
        stabilizer_operate_qgate(self, PHASE_SHIFT_S_, q, 0)
        return self
    
    # 2-qubit gate

    def cx(self, q0, q1):
        """
        operate CX (CNOT) gate.

        Parameters
        ----------
        q : int
            qubit id.

        Returns
        -------
        self : instans of Stabilizer

        """
        stabilizer_operate_qgate(self, CONTROLLED_X, q0, q1)
        return self
    
    def cz(self, q0, q1):
        """
        operate CZ gate.

        Parameters
        ----------
        q : int
            qubit id.

        Returns
        -------
        self : instans of Stabilizer

        """
        stabilizer_operate_qgate(self, CONTROLLED_Z, q0, q1)
        return self
    
    def cy(self, q0, q1):
        """
        operate CY gate.

        Parameters
        ----------
        q : int
            qubit id.

        Returns
        -------
        self : instans of Stabilizer

        """
        self.cz(q0, q1).cx(q0, q1).s(q0)
        return self

    # measurement

    def measure(self, qid=None):
        """
        one shot measurement in Z-direction.

        Parameters
        ----------
        qid : list of int
            qubit id list to measure.

        Returns
        -------
        mval : str
            measurement value.

        Examples
        --------
        >>> sb = Stabilizer(qubit_num=2).set_all('Z').h(0).cx(0,1)
        >>> sb.show()
        >>> print(sb.measure(qid=[0,1]))
        >>> sb.show()
        g[0]:  XX
        g[1]:  ZZ
        00
        g[0]:  ZI
        g[1]:  ZZ

        """
        mval = self.m(qid=qid, shots=1).last
        return mval
    
    def m(self, qid=None, shots=DEF_SHOTS):
        """
        measurement in Z-direction.

        Parameters
        ----------
        qid : list of int
            qubit id list to measure.
        shots : int, default 1
            number of measurements.

        Returns
        -------
        md : instance of MDataStabilizer
            measurement data.

        Examples
        --------
        >>> sb = Stabilizer(qubit_num=2).set_all('Z').h(0).cx(0,1)
        >>> md = qs.m(qid=[0,1], shots=100)
        >>> print(md.freauency)
        >>> print(md.last)
        Counter({'00':55,'11':45})
        11

        See Also
        --------
        MDataStabilizer class

        """
        md = self.mz(qid=qid, shots=shots)
        return md
    
    def mx(self, qid=None, shots=DEF_SHOTS):
        """
        measurement in X-direction.

        Parameters
        ----------
        qid : list of int
            qubit id list to measure.
        shots : int, default 1
            number of measurements.

        Returns
        -------
        md : instance of MDataStabilizer
            measurement data.

        """
        [self.h(q) for q in qid]
        md = self.mz(qid=qid, shots=shots)
        [self.h(q) for q in qid]
        return md
    
    def my(self, qid=None, shots=DEF_SHOTS):
        """
        measurement in Y-direction.

        Parameters
        ----------
        qid : list of int
            qubit id list to measure.
        shots : int, default 1
            number of measurements.

        Returns
        -------
        md : instance of MDataStabilizer
            measurement data.

        """

        [self.s_dg(q).h(q) for q in qid]
        md = self.mz(qid=qid, shots=shots)
        [self.h(q).s(q) for q in qid]
        return md
    
    def mz(self, qid=None, shots=DEF_SHOTS):
        """
        measurement in Z-direction.

        Parameters
        ----------
        qid : list of int
            qubit id list to measure.
        shots : int, default 1
            number of measurements.

        Returns
        -------
        md : instance of MDataStabilizer
            measurement data.

        Examples
        --------
        >>> sb = Stabilizer(2).set_all('Z').h(0).cx(0,1)
        >>> md = qs.mz(qid=[0,1], shots=100)
        >>> print(md.freauency)
        >>> print(md.last)
        Counter({'00':55,'11':45})
        11

        See Also
        --------
        MDataStabilizer class

        """
        frequency = Counter()

        # 1st to last-1 measurement
        for _ in range(shots-1):
            st = self.clone()
            mval_str = ''
            for i in range(len(qid)):
                mval = stabilizer_measure(st, qid[i])
                mval_str += str(mval)
            frequency += {mval_str:1}

        # last measurement
        mval_str = ''
        for i in range(len(qid)):
            m = stabilizer_measure(self, qid[i])
            mval_str += str(m)
        frequency += {mval_str:1}
        last = mval_str

        md = MDataStabilizer(frequency=frequency, last=last, qid=qid,
                             qubit_num=self.qubit_num)

        return md

    @classmethod
    def free_all(cls, *stabs):
        """
        free memory of the all stabilizer.

        Parameters
        ----------
        stabs : instance of Stabilizer,instance of Stabilizer,...
            set of Stabilizer instances

        Returns
        -------
        None

        """
        warnings.warn("No need to call 'free_all' method because free automatically, or you can use class method 'del_all' to free memory explicitly.")

    @classmethod
    def del_all(cls, *stabs):
        """
        free memory of the all stabilizer.

        Parameters
        ----------
        stabs : instance of Stabilizer,instance of Stabilizer,...
            set of Stabilizer instances

        Returns
        -------
        None

        """
        for sb in stabs:
            if type(sb) is list or type(sb) is tuple:
                cls.del_all(*sb)
            elif type(sb) is Stabilizer:
                del sb
            else:
                raise Stabilizer_Error_FreeAll()

    def operate(self, pp=None, ctrl=None):
        """
        operate unitary operator to stabilizer.

        Parameters
        ----------
        pp : instance of PauliProduct
            pauli product to operate
        ctrl : int
            contoroll qubit id for controlled pauli product

        Returns
        -------
        self : instance of Stabilizer
            stabilizer after operation

        """
        pauli_list = pp.pauli_list
        qid = pp.qid
    
        if ctrl is None:
            for q, pauli in zip(qid, pauli_list):
                if pauli == 'X':
                    self.x(q)
                elif pauli == 'Y':
                    self.y(q)
                elif pauli == 'Z':
                    self.z(q)
                else:
                    continue
        else:
            if ctrl in qid:
                raise ValueError("controll and target qubit id conflict")
        
            for q, pauli in zip(qid, pauli_list):
                if pauli == 'X':
                    self.cx(ctrl, q)
                elif pauli == 'Y':
                    self.cy(ctrl, q)
                elif pauli == 'Z':
                    self.cz(ctrl, q)
                else:
                    continue
    
        return self

    def free(self):
        """
        free memory of stabilizer.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        warnings.warn("No need to call 'free' method because free automatically, or you can use 'del' to free memory explicitly.")

    def __del__(self):

        stabilizer_free(self)
        
# c-library for stabilizer
from qlazy.lib.stabilizer_c import *



# -*- coding: utf-8 -*-
from collections import Counter

from qlazy.error import *
from qlazy.config import *
from qlazy.util import *
from qlazy.Stabilizer import *
from qlazy.Result import *
from qlazy.lib.stabilizer_c import *

def run(qcirc=None, shots=1, cid=[], backend=None):

    if qcirc is None:
        raise ValueError("quantum circuit must be specified.")
    
    qubit_num = qcirc.qubit_num
    cmem_num = qcirc.cmem_num
    stab = Stabilizer(qubit_num)
    stab.set_all('Z')
    if cmem_num > 0:
        cmem = CMem(cmem_num)
    else:
        cmem = None
    
    if cmem_num < len(cid):
        raise ValueError("length of cid must be less than classical resister size of qcirc")

    if cid == []:
        cid = [i for i in range(cmem_num)]
        
    qcirc_unitary, qcirc_non_unitary = qcirc.split_unitary_non_unitary()

    frequency = stabilizer_operate_qcirc(stab, cmem, qcirc_unitary, 1, cid)
    frequency = stabilizer_operate_qcirc(stab, cmem, qcirc_non_unitary, shots, cid)

    info = {'stabilizer': stab, 'cmem': cmem}

    result = Result(cid=cid, frequency=frequency, backend=backend, info=info)
    
    return result



# -*- coding: utf-8 -*-
import ctypes
from ctypes.util import find_library
import pathlib

from qlazy.config import *
from qlazy.error import *
from qlazy.util import *
from qlazy.Stabilizer import Stabilizer
from qlazy.MData import *
from qlazy.QCirc import QCirc
from qlazy.CMem import CMem
from qlazy.lib.mdata_c import *

lib= ctypes.CDLL(str(pathlib.Path(__file__).with_name('libqlz.'+get_lib_ext())))
libc = ctypes.CDLL(find_library("c"),mode=ctypes.RTLD_GLOBAL)

def stabilizer_init(gene_num=None, qubit_num=None, seed=None):

    libc.srand(ctypes.c_int(seed))
    
    stab = None
    c_stab = ctypes.c_void_p(stab)
        
    lib.stabilizer_init.restype = ctypes.c_int
    lib.stabilizer_init.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_int,
                                    ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.stabilizer_init(ctypes.c_int(gene_num), ctypes.c_int(qubit_num),
                              ctypes.c_int(seed), c_stab)

    if ret == FALSE:
        raise Stabilizer_Error_Initialize()
        
    return c_stab

def stabilizer_copy(sb):

    try:
        stab = None
        c_stab = ctypes.c_void_p(stab)
            
        lib.stabilizer_copy.restype = ctypes.c_int
        lib.stabilizer_copy.argtypes = [ctypes.POINTER(Stabilizer),
                                        ctypes.POINTER(ctypes.c_void_p)]
        ret = lib.stabilizer_copy(ctypes.byref(sb), c_stab)

        if ret == FALSE:
            raise Stabilizer_Error_Clone()

        return c_stab
        
    except Exception:
        raise Stabilizer_Error_Clone()

def stabilizer_set_pauli_fac(sb, gene_id, pauli_fac):

    lib.stabilizer_set_pauli_fac.restype = ctypes.c_int
    lib.stabilizer_set_pauli_fac.argtypes = [ctypes.POINTER(Stabilizer), ctypes.c_int,
                                             ctypes.c_int]
    ret = lib.stabilizer_set_pauli_fac(ctypes.byref(sb), ctypes.c_int(gene_id),
                                       ctypes.c_int(pauli_fac))

    if ret == FALSE:
        raise Stabilizer_Error_SetPauliFac()

    return ret
    
def stabilizer_get_pauli_fac(sb, gene_id):

    pauli_fac = REAL_PLUS
    c_pauli_fac = ctypes.c_int(pauli_fac)

    lib.stabilizer_get_pauli_fac.restype = ctypes.c_int
    lib.stabilizer_get_pauli_fac.argtypes = [ctypes.POINTER(Stabilizer), ctypes.c_int,
                                             ctypes.POINTER(ctypes.c_int)]
    ret = lib.stabilizer_get_pauli_fac(ctypes.byref(sb), ctypes.c_int(gene_id),
                                       ctypes.byref(c_pauli_fac))

    if ret == FALSE:
        raise Stabilizer_Error_GetPauliFac()

    pauli_fac = c_pauli_fac.value

    return pauli_fac

def stabilizer_set_pauli_op(sb, gene_id=None, qubit_id=None, pauli_op=None):

    lib.stabilizer_set_pauli_op.restype = ctypes.c_int
    lib.stabilizer_set_pauli_op.argtypes = [ctypes.POINTER(Stabilizer), ctypes.c_int,
                                            ctypes.c_int, ctypes.c_int]
    ret = lib.stabilizer_set_pauli_op(ctypes.byref(sb), ctypes.c_int(gene_id),
                                      ctypes.c_int(qubit_id), ctypes.c_int(pauli_op))

    if ret == FALSE:
        raise Stabilizer_Error_SetPauliOp()

    return ret

def stabilizer_get_pauli_op(sb, gene_id=None, qubit_id=None):

    pauli_op = IDENTITY
    c_pauli_op = ctypes.c_int(pauli_op)

    lib.stabilizer_get_pauli_op.restype = ctypes.c_int
    lib.stabilizer_get_pauli_op.argtypes = [ctypes.POINTER(Stabilizer), ctypes.c_int,
                                            ctypes.c_int, ctypes.POINTER(ctypes.c_int)]
    ret = lib.stabilizer_get_pauli_op(ctypes.byref(sb), ctypes.c_int(gene_id),
                                      ctypes.c_int(qubit_id), ctypes.byref(c_pauli_op))

    if ret == FALSE:
        raise Stabilizer_Error_GetPauliOp()

    pauli_op = c_pauli_op.value

    return pauli_op

def stabilizer_operate_qgate(sb, kind=None, q0=None, q1=None):

    lib.stabilizer_operate_qgate.restype = ctypes.c_int
    lib.stabilizer_operate_qgate.argtypes = [ctypes.POINTER(Stabilizer), ctypes.c_int,
                                             ctypes.c_int, ctypes.c_int]
    ret = lib.stabilizer_operate_qgate(ctypes.byref(sb), ctypes.c_int(kind),
                                       ctypes.c_int(q0), ctypes.c_int(q1))
    
    if ret == FALSE:
        raise Stabilizer_Error_OperateQgate()
    
def stabilizer_get_rank(sb):

    rank = 0
    c_rank = ctypes.c_int(rank)

    lib.stabilizer_get_rank.restype = ctypes.c_int
    lib.stabilizer_get_rank.argtypes = [ctypes.POINTER(Stabilizer),
                                        ctypes.POINTER(ctypes.c_int)]
    ret = lib.stabilizer_get_rank(ctypes.byref(sb), ctypes.byref(c_rank))
    
    if ret == FALSE:
        raise Stabilizer_Error_GetRank()

    rank = c_rank.value

    return rank
    
def stabilizer_measure(sb, q=None):

    prob = [0.0, 0.0]
    DoubleArray = ctypes.c_double * 2
    c_prob = DoubleArray(*prob)

    mval = 0
    c_mval = ctypes.c_int(mval)

    lib.stabilizer_measure.restype = ctypes.c_int
    lib.stabilizer_measure.argtypes = [ctypes.POINTER(Stabilizer), ctypes.c_int,
                                       DoubleArray, ctypes.POINTER(ctypes.c_int)]
    ret = lib.stabilizer_measure(ctypes.byref(sb), ctypes.c_int(q),
                                 c_prob, ctypes.byref(c_mval))
    
    if ret == FALSE:
        raise Stabilizer_Error_Measure()

    mval = c_mval.value

    return mval
    
def stabilizer_operate_qcirc(sb, cmem, qcirc, shots, cid):

    lib.stabilizer_operate_qcirc.restype = ctypes.c_int
    lib.stabilizer_operate_qcirc.argtypes = [ctypes.POINTER(Stabilizer), ctypes.POINTER(CMem), ctypes.POINTER(QCirc)]

    if cmem is not None:

        cmem_num = cmem.cmem_num
        frequency = Counter()
        for n in range(shots):
            
            if n < shots - 1:
                sb_tmp = sb.clone()
                cmem_tmp = cmem.clone()
                ret = lib.stabilizer_operate_qcirc(ctypes.byref(sb_tmp), ctypes.byref(cmem_tmp), ctypes.byref(qcirc))
                bit_array = ctypes.cast(cmem_tmp.bit_array, ctypes.POINTER(ctypes.c_int*cmem_num))
            else:
                ret = lib.stabilizer_operate_qcirc(ctypes.byref(sb), ctypes.byref(cmem), ctypes.byref(qcirc))
                bit_array = ctypes.cast(cmem.bit_array, ctypes.POINTER(ctypes.c_int*cmem_num))

            if ret == FALSE:
                raise Stabilizer_Error_OperateQcirc()
        
            cmem_list = [bit_array.contents[i] for i in range(cmem_num)]
            cmem_list_part = [cmem_list[c] for c in cid]
            mval = "".join(map(str, cmem_list_part))

            frequency[mval] += 1

        return frequency
    
    else:

        c_cmem = ctypes.POINTER(CMem)()
        ret = lib.stabilizer_operate_qcirc(ctypes.byref(sb), c_cmem, ctypes.byref(qcirc))
        
        if ret == FALSE:
            raise Stabilizer_Error_OperateQcirc()

        return None
    
def stabilizer_free(stab):

    lib.stabilizer_free.argtypes = [ctypes.POINTER(Stabilizer)]
    lib.stabilizer_free(ctypes.byref(stab))



# -*- coding: utf-8 -*-

from qlazy.UnitaryOperator import UnitaryOperator
        
class PauliProduct(UnitaryOperator):

    """ Pauli Product

    Attributes
    ----------
    qid : list of int
        list of qubit id.
    pauli_list : list of str
        list of pauli operators.

    """
    def __init__(self, pauli_str=None, qid=None):
        """
        Parameters
        ----------
        pauli_str : str
            string representation of Pauli Product.
            ex) "XZZY"
        qid : list of int, default - [0,1,..,len(pauli_str)-1]
            list of qubit id.
            ex) [1, 0, 3, 6]

        Notes
        -----
        'pauli_str' must consist of 'X','Y','Z', and length of 'pauli_str' must be equal to length of 'qid'

        Examples
        --------
        >>> pp = PauliProduct(pauli_str ="XZZY", qid=[1, 0, 3, 6])
        represent Pauli Product : X1*Z0*Z3*Y6.
        >>> pp = PauliProduct(pauli_str ="I")
        represent identity.

        """
        if pauli_str is None:
            raise ValueError("pauli_str or qid is not specified")
        if qid is None:
            qid = [i for i in range(len(pauli_str))]
    
        if len(pauli_str) != len(qid):
            raise ValueError("length of 'pauli_str' is not equal to length of 'qid'")
        elif len(qid) != len(set(qid)):
            raise ValueError("some elements of 'qid' are duplicated")
        elif sum([0 if p in ['X','Y','Z','I'] else 1 for p in list(pauli_str)]) > 0:
            raise ValueError("'pauli_str' must consist of 'X','Y','Z'")

        qid_sorted = []
        pauli_list_sorted = []

        if list(set(pauli_str)) == ['I']:
            qid_sorted = [0]
            pauli_list_sorted = ['I']
        else:
            for q, p in sorted(zip(qid, list(pauli_str))):
                if p == 'I': continue
                qid_sorted.append(q)
                pauli_list_sorted.append(p)

        super().__init__(qid=qid_sorted)
        self.pauli_list = pauli_list_sorted

    def __str__(self):

        pauli_product = ""
        for i,p in sorted(zip(self.qid, self.pauli_list)):
            pauli_product += "{0:}({1:}) ".format(p, str(i))

        pauli_product = pauli_product.rstrip()
        return pauli_product

    def get_binary_rep(self, qubit_num=None):
        """
        get binary representation of pauli product.

        Parameters
        ----------
        qubit_num : int, default - max(self.qid)+1
            qubit number of binary representation.

        Returns
        -------
        binary_rep : list of int
            output binary representation of the pauli product.

        Examples
        --------
        >>> pp = PauliProduct(pauli_str="XYZ", qid=[0,1,2])
        >>> print(pp.get_binary_rep())
        [1,1,0,0,1,1]

        """
        if qubit_num is None:
            qubit_num = self.qubit_num

        if qubit_num < max(self.qid) + 1:
            raise Valueerror("qubit_num is smaller than the size required by pauli product")
            
        binary_rep = [0] * qubit_num * 2
        for q in range(qubit_num):
            if q in self.qid:
                i = self.qid.index(q)
                if self.pauli_list[i] == 'X':
                    binary_rep[q] = 1
                elif self.pauli_list[i] == 'Z':
                    binary_rep[qubit_num + q] = 1
                elif self.pauli_list[i] == 'Y':
                    binary_rep[q] = 1
                    binary_rep[qubit_num + q] = 1

        return binary_rep

    def is_commute(self, pp):
        """
        is pauli product commute to other pauli product or not.

        Parameters
        ----------
        pp : instance of PauliProduct
            pauli product

        Returns
        -------
        ans : bool
            commute (True) or anti-commute (False)

        """
        qubit_num = max(max(self.qid), max(pp.qid)) + 1

        binary_rep_self = self.get_binary_rep(qubit_num)
        binary_rep_pp = pp.get_binary_rep(qubit_num)

        inner_prod = 0
        for i in range(qubit_num):
            inner_prod += (binary_rep_self[i] * binary_rep_pp[qubit_num + i])
            inner_prod += (binary_rep_self[qubit_num + i] * binary_rep_pp[i])
        inner_prod = inner_prod % 2

        if inner_prod == 0:
            ans = True
        else:
            ans = False
        
        return ans

    def tenspro(self, pp):
        """
        get tensor product with other pauli product.

        Parameters
        ----------
        pp : instance of PauliProduct
            pauli product

        Returns
        -------
        pp_out : instance of PauliProduct
            output pauli product

        """
        qubit_num = max(max(self.qid), max(pp.qid)) + 1
        
        binary_rep_self = self.get_binary_rep(qubit_num)
        binary_rep_pp = pp.get_binary_rep(qubit_num)

        binary_rep = [(bs + bp) % 2 for bs, bp in zip(binary_rep_self, binary_rep_pp)]

        pauli_list = []
        qid = []
        for i in range(qubit_num):
            if binary_rep[i] == 1 and binary_rep[qubit_num + i] == 1:
                pauli_list.append('Y')
                qid.append(i)
            elif binary_rep[i] == 1:
                pauli_list.append('X')
                qid.append(i)
            elif binary_rep[qubit_num + i] == 1:
                pauli_list.append('Z')
                qid.append(i)

        pauli_str = ''.join(pauli_list)

        pp_out = __class__(pauli_str=pauli_str, qid=qid)

        return pp_out



# -*- coding: utf-8 -*-
import ctypes
from ctypes.util import find_library
import pathlib
from collections import Counter

from qlazy.config import *
from qlazy.error import *
from qlazy.util import *

class MData(ctypes.Structure):

    _fields_ = [
        ('qubit_num', ctypes.c_int),
        ('shot_num', ctypes.c_int),
        ('angle', ctypes.c_double),
        ('phase', ctypes.c_double),
        ('qubit_id', ctypes.POINTER(ctypes.c_int)),
        ('freq', ctypes.POINTER(ctypes.c_int)),
        ('last_val', ctypes.c_int),
    ]

    def __new__(cls, qubit_num=None, shots=1, angle=0.0, phase=0.0, qid=[], **kwargs):
        """
        Parameters
        ----------
        None

        Returns
        -------
        md : instance (QCirc)

        """
        obj = mdata_init(qubit_num, shots, angle, phase, qid)
        md = ctypes.cast(obj.value, ctypes.POINTER(cls)).contents
        return md

    def __str__(self):

        s = ""
        s += "measured direction (theta, phi): {0:}*PI, {1:}*PI\n".format(self.angle, self.phase)
        s += "measured qubit id: {}\n".format([self.qubit_id[i] for i in range(self.qubit_num)])
        s += "frequency: {}\n".format(self.frequency)
        s += "last state: {}".format(self.last)
        return s
    
    def show(self):

        mdata_print(self)

    @property
    def last(self):
        """ last measured value (binary string) """
        mval = self.last_val
        digits = self.qubit_num
        return '{:0{digits}b}'.format(mval, digits=digits)
            
    @property
    def lst(self):
        """ last measured value (integer) """
        return self.last_val
            
    @property
    def frequency(self):
        """ frequencies of measured value (Counter) """
        state_num = 2**self.qubit_num
        frq = [self.freq[i] for i in range(state_num)]
        digits = self.qubit_num
        res = {"{:0{digits}b}".format(k, digits=digits):v
               for k,v in enumerate(frq) if v > 0}
        return Counter(res)
            
    @property
    def frq(self):
        """ frequencies of measured value (list) """
        state_num = 2**self.qubit_num
        freq_list = [0] * state_num
        for k, v in self.frequency.items():
            i = int(k, 2)
            freq_list[i] = v
        return freq_list

    @property
    def state_num(self):
        return 2**self.qubit_num
        
    def __del__(self):

        mdata_free(self)
        
# c-library for qstate
from qlazy.lib.mdata_c import *



# -*- coding: utf-8 -*-
import ctypes
from ctypes.util import find_library
import numpy as np
import pathlib

from qlazy.config import *
from qlazy.error import *
from qlazy.util import *
from qlazy.CMem import CMem

lib= ctypes.CDLL(str(pathlib.Path(__file__).with_name('libqlz.'+get_lib_ext())))
libc = ctypes.CDLL(find_library("c"),mode=ctypes.RTLD_GLOBAL)

def cmem_init(cmem_num):

    if cmem_num < 1:
        raise CMem_Error_Initialize()
    
    cmem = None
    c_cmem = ctypes.c_void_p(cmem)

    lib.cmem_init.restype = ctypes.c_int
    lib.cmem_init.argtypes = [ctypes.c_int, ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.cmem_init(ctypes.c_int(cmem_num), c_cmem)

    if ret == FALSE:
        raise CMem_Error_Initialize()

    return c_cmem

def cmem_copy(cm):
    
    cmem = None
    c_cmem = ctypes.c_void_p(cmem)
            
    lib.cmem_copy.restype = ctypes.c_int
    lib.cmem_copy.argtypes = [ctypes.POINTER(CMem),
                              ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.cmem_copy(ctypes.byref(cm), c_cmem)

    if ret == FALSE:
        raise CMem_Error_Clone()

    return c_cmem

def cmem_free(cmem):

    lib.cmem_free.argtypes = [ctypes.POINTER(CMem)]
    lib.cmem_free(ctypes.byref(cmem))



# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod, ABC

class UnitaryOperator(metaclass=ABCMeta):

    """ Unitary Operator

    Attributes
    ----------
    qid : list of int
        list of qubit id.

    """
    @abstractmethod
    def __init__(self, qid=None):

        if qid is None:
            raise ValueError("'qid' is not specified")
        
        self.qid = qid
        self.qubit_num = max(qid) + 1

    @abstractmethod
    def __str__(self):
        raise NotImplemented()

    @abstractmethod
    def tenspro(self, unitary_operator):
        raise NotImplemented()



# -*- coding: utf-8 -*-
from qlazy.config import *
from qlazy.error import *

BACKEND_DEVICES = {'qlazy': ['qstate_simulator','stabilizer_simulator'],
                   'qulacs': ['cpu_simulator','gpu_simulator'],
                   'ibmq': ['aer_simulator', 'qasm_simulator']}

class Backend:
    """ Backend device of quantum computing

    Attributes
    ----------
    product : str
        product name of the backend
        ('qlazy' or 'qulacs' or 'ibmq')
    device : str
        device name of the product

    Example
    -------
    >>> bk = Backend(product='qlazy', device='stabilizer_simulator')
    >>> bk = Backend(product='qulacs', device='gpu_simulator')
    >>> bk = Backend(product='ibmq', device='least_busy')

    """

    def __init__(self, product=None, device=None):

        #
        # set attributes (procuct, device)
        #
        
        if product == None:
            if device == None:
                product = 'qlazy'
                device = 'qstate_simulator'
            else:
                raise Backend_Error_NameNotSpecified()            
        
        if product in BACKEND_DEVICES.keys():
            self.product = product

            if device in BACKEND_DEVICES[product]:
                self.device = device
            elif device == None:
                self.device = BACKEND_DEVICES[product][0]
            else:
                if product == 'ibmq':
                    self.device = device
                else:
                    raise Backend_Error_DeviceNotSupported()

        else:
            raise Backend_Error_NameNotSupported()

        #
        # set method (__run)
        #
        
        # qlazy
        if self.product == 'qlazy':
    
            # qstate simulator
            if self.device == 'qstate_simulator':
                from qlazy.backend.qlazy_qstate_simulator import run
                self.__run = run
        
            # stabilizer simulator
            elif self.device == 'stabilizer_simulator':
                from qlazy.backend.qlazy_stabilizer_simulator import run
                self.__run = run
                
            else:
                raise Backend_Error_DeviceNotSupported()
            
        # qulacs
        elif self.product == 'qulacs':
    
            # cpu_simulator
            if self.device == 'cpu_simulator':
                from qlazy.backend.qulacs_simulator import run_cpu
                self.__run = run_cpu
        
            # gpu_simulator
            elif self.device == 'gpu_simulator':
                from qlazy.backend.qulacs_simulator import run_gpu
                self.__run = run_gpu
                
            else:
                raise Backend_Error_DeviceNotSupported()
        
        # ibmq
        elif self.product == 'ibmq':
            from qlazy.backend.ibmq import run
            self.__run = run
    
        else:
            raise Backend_Error_NameNotSupported()

    @classmethod
    def products(cls):

        return list(BACKEND_DEVICES)

    @classmethod
    def qlazy_devices(cls):
        return BACKEND_DEVICES['qlazy']
    
    @classmethod
    def qulacs_devices(cls):
        return BACKEND_DEVICES['qulacs']
    
    @classmethod
    def ibmq_devices(cls):
        devices = BACKEND_DEVICES['ibmq']
        try:
            from qiskit import IBMQ
            provider = IBMQ.load_account()
            ibmq_backend_system_names = [b.name() for b in
                                         provider.backends(simulator=False, operational=True)]
        except Exception as e:
            print(e)
        else:
            devices += ['least_busy']
            devices += ibmq_backend_system_names
        
        return devices

    @classmethod
    def devices(cls, product):

        if product == 'qlazy':
            return Backend.qlazy_devices()
        elif product == 'qulacs':
            return Backend.qulacs_devices()
        elif product == 'ibmq':
            return Backend.ibmq_devices()
        else:
            raise ValueError("unknown product: {}".format(product))

    def run(self, qcirc=None, shots=1, cid=[]):
        """
        run the quantum circuit.

        Parameters
        ----------
        qcirc : instance of QCirc, default None
            quantum circuit.
        shots : int, default 1
            number of measurements.
        cid : list, default []
            classical register id list to count frequency.

        Returns
        -------
        result : incetance of Result
            measurement result.

        Examples
        --------
        >>> # simple example
        >>> from qlazy import QCirc, Backend
        >>> bk = Backend(product='qlazy', device='qstate_simulator')
        >>> qc = QCirc().h(0).cx(0,1).measure(qid=[0,1], cid=[0,1])
        >>> result = bk.run(qcirc=qc, shots=100)
        >>> print(result.frequency)
        Counter({'00': 51, '11': 49})
        >>> ...
        >>> # control quantum gate by measured results
        >>> from qlazy import QCirc, Backend
        >>> bk = Backend(product='qlazy', device='qstate_simulator')
        >>> qc = QCirc.h(0).cx(0,1).measure(qid=[0],cid=[0]).x(0, ctrl=0).x(1, ctrl=0).measure(qid=[0,1], cid=[0,1])
        >>> result = bk.run(qcirc=qc, shots=100)
	>>> print(result.frequency)
        Counter({'00': 100})

        """
        result = self.__run(qcirc=qcirc, shots=shots, cid=cid, backend=self)
        return result



# -*- coding: utf-8 -*-
import ctypes
import warnings
import ctypes

from qlazy.config import *
from qlazy.error import *

class Observable(ctypes.Structure):
    """ Obserbavle for quantum many-body spin system. """
    
    _fields_ = [
        ('spin_num', ctypes.c_int),
        ('array_num', ctypes.c_int),
        ('spro_array', ctypes.c_void_p),
    ]

    def __new__(cls, string=None, **kwargs):
        """
        Parameters
        ----------
        string : str
            expression for the observable.

        Returns
        -------
        ob : instance of Observable
            observable corresponding to the string expression.

        Examles
        -------
        Hamiltonian for 2-body spin system : -2.0 + Z0 X1 + X0 X1
        >>> ob = Observable("-2.0+z_0*z_1+x_0+x_1")

        """
        cls.spin_num = 0
        cls.array_num = 0
        cls.spro_array = None
        obj = observable_init(string)
        ob = ctypes.cast(obj.value, ctypes.POINTER(cls)).contents
        # ob = observable_init(string)
        return ob

    def free(ob):
        """
        free memory of observable.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        # observable_free(ob)
        warnings.warn("No need to call 'free' method because free automatically, or you can use 'del' to free memory explicitly.")

    def __del__(self):
        
        observable_free(self)

# c-library for observable
from qlazy.lib.observable_c import *



# -*- coding: utf-8 -*-
import sys
import random
from collections import Counter
import numpy as np
import cmath

from qlazy.error import *
from qlazy.config import *
from qlazy.util import *
from qlazy.Result import *

from qiskit import IBMQ, QuantumCircuit, ClassicalRegister, QuantumRegister
from qiskit import execute, Aer
from qiskit.qasm import pi
from qiskit.providers.ibmq import least_busy
from qiskit.circuit.library.standard_gates import SXdgGate

def run(qcirc=None, shots=1, cid=[], backend=None):
            
    if qcirc is None:
        raise ValueError("quantum circuit must be specified.")

    qcirc_tmp = qcirc.clone()
    
    qubit_num = qcirc_tmp.qubit_num
    cmem_num = qcirc_tmp.cmem_num

    if cmem_num < len(cid):
        raise ValueError("length of cid must be less than classical resister size of qcirc")
    
    if cid == []:
        cid = [i for i in range(cmem_num)]

    qubit_reg = QuantumRegister(qubit_num)
    cmem_reg = [ClassicalRegister(1, name="c{}".format(i)) for i in range(cmem_num)]

    args = [qubit_reg] + cmem_reg
    qc = QuantumCircuit(*args)

    exist_measurement = False
    while True:
        kind = qcirc_tmp.kind_first()
        if kind == None: break

        (kind, qid, para, c, ctrl) = qcirc_tmp.pop_gate()

        if kind == MEASURE:
            exist_measurement = True
            if c == None:
                raise ValueError("cid (classical register ID) must be specified")
            else:
                qc.measure(qubit_reg[qid[0]], cmem_reg[c])
                
        elif kind == RESET:
            qc.reset(qubit_reg[qid[0]])
            
        else:
            __ibmq_add_qgate(qc, kind, qid, para[0] * np.pi, para[1] * np.pi, para[2] * np.pi,
                             ctrl, cmem_reg)

    # set backend
    if backend.device == 'aer_simulator':
        ibmq_backend = Aer.get_backend("aer_simulator")
    elif backend.device == 'qasm_simulator':
        ibmq_backend = Aer.get_backend("qasm_simulator")
    else:
        provider = IBMQ.load_account()
        if backend.device == 'least_busy':
            ibmq_backend = least_busy(provider.backends(simulator=False, operational=True))
        else:
            ibmq_backend_system_names = [b.name() for b in provider.backends(simulator=False, operational=True)]
            if backend.device in ibmq_backend_system_names:
                ibmq_backend = provider.get_backend(backend.device)
            else:
                raise ValueError("unknown device")

    # execute the circuit
    if exist_measurement == True:
        res = execute(qc, ibmq_backend, shots=shots).result()
        frq = res.get_counts(qc)

        frequency = Counter()
        for k,v in frq.items():
            bits_list = list(k.replace(' ', ''))
            bits_list.reverse()
            measured_bits_list = [bits_list[c] for c in cid]
            measured_bits = ''.join(measured_bits_list)
            if measured_bits != '':
                frequency[measured_bits] += v
            else:
                frequency = None
                break
    else:  # no measurement gates included
        cid = []
        frequency = None

    # execute the circuit (for state vector)
    ibmq_sv_backend = Aer.get_backend("statevector_simulator")
    res_sv = execute(qc, ibmq_sv_backend).result()
    statevector = np.asarray(res_sv.get_statevector(qc))

    info = {'statevector': statevector, 'creg': cmem_reg, 'quantum_circuit': qc}
    result = Result(cid=cid, frequency=frequency, backend=backend, info=info)

    return result

def __ibmq_add_qgate(qc, kind, qid, phase, phase1, phase2, ctrl, cmem_reg):

    # 1-qubit, 0-parameter gate
    if kind == PAULI_X:
        if ctrl == None: qc.x(qid[0])
        else: qc.x(qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == PAULI_Y:
        if ctrl == None: qc.y(qid[0])
        else: qc.y(qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == PAULI_Z:
        if ctrl == None: qc.z(qid[0])
        else: qc.z(qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == ROOT_PAULI_X:
        if ctrl == None: qc.sx(qid[0])
        else: qc.sx(qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == ROOT_PAULI_X_:
        if ctrl == None: qc.sxdg(qid[0])
        else: qc.sxdg(qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == HADAMARD:
        if ctrl == None: qc.h(qid[0])
        else: qc.h(qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == PHASE_SHIFT_S:
        if ctrl == None: qc.s(qid[0])
        else: qc.s(qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == PHASE_SHIFT_S_:
        if ctrl == None: qc.sdg(qid[0])
        else: qc.sdg(qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == PHASE_SHIFT_T:
        if ctrl == None: qc.t(qid[0])
        else: qc.t(qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == PHASE_SHIFT_T_:
        if ctrl == None: qc.tdg(qid[0])
        else: qc.tdg(qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == IDENTITY:
        if ctrl == None: qc.i(qid[0])
        else: qc.i(qid[0]).c_if(cmem_reg[ctrl], 1)
        
    # 1-qubit, 1-parameter gate
    elif kind == ROTATION_X:
        if ctrl == None: qc.rx(phase, qid[0])
        else: qc.rx(phase, qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == ROTATION_Y:
        if ctrl == None: qc.ry(phase, qid[0])
        else: qc.ry(phase, qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == ROTATION_Z:
        if ctrl == None: qc.rz(phase, qid[0])
        else: qc.rz(phase, qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == PHASE_SHIFT:
        if ctrl == None: qc.p(phase, qid[0])
        else: qc.p(phase, qid[0]).c_if(cmem_reg[ctrl], 1)

    elif kind == ROTATION_U1:
        if ctrl == None: qc.p(phase, qid[0])
        else: qc.p(phase, qid[0]).c_if(cmem_reg[ctrl], 1)
        
    # 1-qubit, 2-parameter gate
    elif kind == ROTATION_U2:
        if ctrl == None: qc.u(0.5 * np.pi, phase1, phase, qid[0])
        else: qc.u(0.5 * np.pi, phase1, phase, qid[0]).c_if(cmem_reg[ctrl], 1)
        
    # 1-qubit, 3-parameter gate
    elif kind == ROTATION_U3:
        if ctrl == None: qc.u(phase2, phase1, phase, qid[0])
        else: qc.u(phase2, phase1, phase, qid[0]).c_if(cmem_reg[ctrl], 1)

    # 2-qubit, 0-parameters gate
    elif kind == CONTROLLED_X:
        if ctrl == None: qc.cx(qid[0], qid[1])
        else: qc.cx(qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_Y:
        if ctrl == None: qc.cy(qid[0], qid[1])
        else: qc.cy(qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_Z:
        if ctrl == None: qc.cz(qid[0], qid[1])
        else: qc.cz(qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_XR:
        if ctrl == None: qc.csx(qid[0], qid[1])
        else: qc.csx(qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_XR_:
        q_csxdg = QuantumRegister(2)
        qc_csxdg = QuantumCircuit(q_csxdg)
        csxdg_gate = SXdgGate().control(1)
        qc_csxdg.append(csxdg_gate, q_csxdg)
        if ctrl == None: qc.append(qc_csxdg, qargs=[qid[0], qid[1]])
        else: qc.append(qc_csxdg, qargs=[qid[0], qid[1]]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_H:
        if ctrl == None: qc.ch(qid[0], qid[1])
        else: qc.ch(qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_S:
        if ctrl == None: qc.cp(0.5 * np.pi, qid[0], qid[1])
        else: qc.cp(0.5 * np.pi, qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_S_:
        if ctrl == None: qc.cp(-0.5 * np.pi, qid[0], qid[1])
        else: qc.cp(-0.5 * np.pi, qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_T:
        if ctrl == None: qc.cp(0.25 * np.pi, qid[0], qid[1])
        else: qc.cp(0.25 * np.pi, qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_T_:
        if ctrl == None: qc.cp(-0.25 * np.pi, qid[0], qid[1])
        else: qc.cp(-0.25 * np.pi, qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_P:
        if ctrl == None: qc.cp(phase, qid[0], qid[1])
        else: qc.cp(phase, qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == SWAP_QUBITS:
        if ctrl == None: qc.swap(qid[0], qid[1])
        else: qc.swap(qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    # 2-qubit, 1-parameters gate
    elif kind == CONTROLLED_RX:
        if ctrl == None: qc.crx(phase, qid[0], qid[1])
        else: qc.crx(phase, qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_RY:
        if ctrl == None: qc.cry(phase, qid[0], qid[1])
        else: qc.cry(phase, qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_RZ:
        if ctrl == None: qc.crz(phase, qid[0], qid[1])
        else: qc.crz(phase, qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    elif kind == CONTROLLED_U1:
        if ctrl == None: qc.cp(phase, qid[0], qid[1])
        else: qc.cp(phase, qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    # 2-qubit, 2-parameters gate
    elif kind == CONTROLLED_U2:
        if ctrl == None: qc.cu(0.5 * np.pi, phase1, phase, 0, qid[0], qid[1])
        else: qc.cu(0.5 * np.pi, phase1, phase, 0, qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    # 2-qubit, 3-parameters gate
    elif kind == CONTROLLED_U3:
        if ctrl == None: qc.cu(phase2, phase1, phase, 0, qid[0], qid[1])
        else: qc.cu(phase2, phase1, phase, 0, qid[0], qid[1]).c_if(cmem_reg[ctrl], 1)

    else:
        raise ValueError("unknown gate")



# -*- coding: utf-8 -*-
import ctypes
from ctypes.util import find_library
import numpy as np
import pathlib

from qlazy.config import *
from qlazy.error import *
from qlazy.util import *
from qlazy.MData import *
from qlazy.QCirc import QCirc
from qlazy.lib.mdata_c import *

lib= ctypes.CDLL(str(pathlib.Path(__file__).with_name('libqlz.'+get_lib_ext())))
libc = ctypes.CDLL(find_library("c"),mode=ctypes.RTLD_GLOBAL)

def qcirc_init():

    qcirc = None
    c_qcirc = ctypes.c_void_p(qcirc)

    lib.qcirc_init.restype = ctypes.c_int
    lib.qcirc_init.argtypes = [ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.qcirc_init(c_qcirc)

    if ret == FALSE:
        raise QCirc_Error_Initialize()

    return c_qcirc

def qcirc_copy(qc):

    qcirc = None
    c_qcirc = ctypes.c_void_p(qcirc)
            
    lib.qcirc_copy.restype = ctypes.c_int
    lib.qcirc_copy.argtypes = [ctypes.POINTER(QCirc),
                               ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.qcirc_copy(ctypes.byref(qc), c_qcirc)

    if ret == FALSE:
        raise QCirc_Error_Copy()

    return c_qcirc

def qcirc_merge(qc_L, qc_R):

    qcirc = None
    c_qcirc = ctypes.c_void_p(qcirc)
            
    lib.qcirc_merge.restype = ctypes.c_int
    lib.qcirc_merge.argtypes = [ctypes.POINTER(QCirc), ctypes.POINTER(QCirc),
                               ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.qcirc_merge(ctypes.byref(qc_L), ctypes.byref(qc_R), c_qcirc)

    if ret == FALSE:
        raise QCirc_Error_Merge()

    return c_qcirc

def qcirc_is_equal(qc_L, qc_R):

    ans = True
    c_ans = ctypes.c_bool(ans)
            
    lib.qcirc_is_equal.restype = ctypes.c_int
    lib.qcirc_is_equal.argtypes = [ctypes.POINTER(QCirc), ctypes.POINTER(QCirc),
                                   ctypes.POINTER(ctypes.c_bool)]
    ret = lib.qcirc_is_equal(ctypes.byref(qc_L), ctypes.byref(qc_R), ctypes.byref(c_ans))

    if ret == FALSE:
        raise QCirc_Error_Merge()

    ans = c_ans.value
    return ans

def qcirc_append_gate(qcirc, kind, qid, para, c, ctrl):

    if para == None: para = [0.0, 0.0, 0.0]
    if c == None: c = -1
    if ctrl == None: ctrl = -1
    
    qid_num = len(qid)
    para_num = len(para)
    IntArray = ctypes.c_int * qid_num
    DoubleArray = ctypes.c_double * para_num
    c_qid = IntArray(*qid)
    c_para = DoubleArray(*para)

    lib.qcirc_append_gate.restype = ctypes.c_int
    lib.qcirc_append_gate.argtypes = [ctypes.POINTER(QCirc), ctypes.c_int, IntArray, DoubleArray,
                                      ctypes.c_int, ctypes.c_int]
    ret = lib.qcirc_append_gate(ctypes.byref(qcirc), ctypes.c_int(kind), c_qid, c_para,
                                ctypes.c_int(c), ctypes.c_int(ctrl))

    if ret == FALSE:
        raise QCirc_Error_AppendGate()
    
def qcirc_kind_first(qc):

    kind = 0
    c_kind = ctypes.c_int(kind)
            
    lib.qcirc_kind_first.restype = ctypes.c_int
    lib.qcirc_kind_first.argtypes = [ctypes.POINTER(QCirc), ctypes.POINTER(ctypes.c_int)]
    ret = lib.qcirc_kind_first(ctypes.byref(qc), ctypes.byref(c_kind))

    if ret == FALSE:
        raise QState_Error_KindFirst()

    kind = c_kind.value

    if kind == NOT_A_GATE:
        kind = None
    
    return kind

def qcirc_pop_gate(qc):

    kind = 0
    c_kind = ctypes.c_int(kind)
    qid = [0] * 2
    IntArray = ctypes.c_int * 2
    c_qid = IntArray(*qid)
    para = [0.0] * 3
    DoubleArray = ctypes.c_double * 3
    c_para = DoubleArray(*para)
    c = -1
    c_c = ctypes.c_int(c)
    ctrl = -1
    c_ctrl = ctypes.c_int(ctrl)
            
    lib.qcirc_pop_gate.restype = ctypes.c_int
    lib.qcirc_pop_gate.argtypes = [ctypes.POINTER(QCirc), ctypes.POINTER(ctypes.c_int),
                                     ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_double),
                                     ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int)]
                                     
    ret = lib.qcirc_pop_gate(ctypes.byref(qc), ctypes.byref(c_kind), c_qid, c_para,
                             ctypes.byref(c_c), ctypes.byref(c_ctrl))

    if ret == FALSE:
        raise QState_Error_PopGate()

    kind = c_kind.value
    qid = [c_qid[i] for i in range(2)]
    para = [c_para[i] for i in range(3)]
    c = c_c.value
    ctrl = c_ctrl.value

    if c == -1: c = None
    if ctrl == -1: ctrl = None

    return (kind, qid, para, c, ctrl)
    
def qcirc_free(qcirc):

    lib.qcirc_free.argtypes = [ctypes.POINTER(QCirc)]
    lib.qcirc_free(ctypes.byref(qcirc))



# -*- coding: utf-8 -*-
from collections import Counter

from qlazy.config import *
from qlazy.error import *

class Result:
    """ Result of quantum circuit execution

    Attributes
    ----------
    cid : list
        classical register id list to get frequencys.
    freqency : Counter
        frequencies of measured value.
    backend : instance of Backend
        backend device of quantum computing
    info : dict
        result information relating to the backend device

    """
    def __init__(self, cid=None, frequency=None, backend=None, info=None):

        self.cid = cid
        self.frequency = frequency
        self.backend = backend
        self.info = info



# -*- coding: utf-8 -*-
import math
import numpy as np
from qlazy.error import *
from qlazy.config import *

def get_lib_ext():
    
    import platform
    if platform.system() == 'Darwin':
        return 'dylib'
    else:
        return 'so'

def qstate_check_args(qs, kind=None, qid=None, shots=None, angle=None,
                      phase=None, phase1=None, phase2=None):
    
    for q in qid:
        if (q >= qs.qubit_num) or (q < 0):
            raise QState_OutOfBound()
            
    qnum = get_qgate_qubit_num(kind)
    
    if qnum == 0:  # any qubit number
        # check qubit number
        if len(qid) > qs.qubit_num:
            raise QState_TooManyArguments()
        elif len(qid) < 1:
            raise QState_NeedMoreArguments()
        else:
            pass
            
        # check same qubit number
        if len(set(qid)) != len(qid):
            raise QState_SameQubitID()
            
    elif qnum == 1:
        # check qubit number
        if len(qid) > qnum:
            raise QState_TooManyArguments()
        elif len(qid) < qnum:
            raise QState_NeedMoreArguments()
        else:
            return True
            
    elif qnum == 2:
        # check qubit number
        if len(qid) > qnum:
            raise QState_TooManyArguments()
        elif len(qid) < qnum:
            raise QState_NeedMoreArguments()
        else:
            pass
    
        # check same qubit number
        if (qid[0]==qid[1]):
            raise QState_SameQubitID()
        else:
            return True
            
    elif qnum == 3:
        # check qubit number
        if len(qid) > qnum:
            raise QState_TooManyArguments()
        elif len(qid) < qnum:
            raise QState_NeedMoreArguments()
        else:
            pass
    
        # check same qubit id
        if (qid[0]==qid[1] or qid[1]==qid[2] or qid[2]==qid[0]):
            raise QState_SameQubitID()
        else:
            return True
    
def densop_check_args(de, kind=None, qid=None, shots=None, angle=None,
                      phase=None, phase1=None, phase2=None):

    qubit_num = int(math.log2(de.row))
        
    for q in qid:
        if (q >= qubit_num) or (q < 0):
            raise QState_OutOfBound()
            
    qnum = get_qgate_qubit_num(kind)

    if qnum == 0:  # any qubit number
        # check qubit number
        if len(qid) > qubit_num:
            raise QState_TooManyArguments()
        elif len(qid) < 1:
            raise QState_NeedMoreArguments()
        else:
            pass
            
        # check same qubit number
        if len(set(qid)) != len(qid):
            raise QState_SameQubitID()
            
    elif qnum == 1:
            # check qubit number
        if len(qid) > qnum:
            raise QState_TooManyArguments()
        elif len(qid) < qnum:
            raise QState_NeedMoreArguments()
        else:
            return True
            
    elif qnum == 2:
        # check qubit number
        if len(qid) > qnum:
            raise QState_TooManyArguments()
        elif len(qid) < qnum:
            raise QState_NeedMoreArguments()
        else:
            pass

        # check same qubit number
        if (qid[0]==qid[1]):
            raise QState_SameQubitID()
        else:
            return True
            
    elif qnum == 3:
        # check qubit number
        if len(qid) > qnum:
            raise QState_TooManyArguments()
        elif len(qid) < qnum:
            raise QState_NeedMoreArguments()
        else:
            pass

        # check same qubit id
        if (qid[0]==qid[1] or qid[1]==qid[2] or qid[2]==qid[0]):
            raise QState_SameQubitID()
        else:
            return True

def get_qgate_qubit_num(kind=None):

    if (kind==SHOW or kind==MEASURE or
        kind==MEASURE_X or kind==MEASURE_Y or kind==MEASURE_Z or kind==RESET):  # 0 if any number
        return 0
    elif (kind==BLOCH or kind==PAULI_X or kind==PAULI_Y or kind==PAULI_Z or
          kind==ROOT_PAULI_X or kind==ROOT_PAULI_X_ or kind==HADAMARD or
          kind==PHASE_SHIFT_S or kind==PHASE_SHIFT_S_ or
          kind==PHASE_SHIFT_T or kind==PHASE_SHIFT_T_ or kind==PHASE_SHIFT or
          kind==ROTATION_X or kind==ROTATION_Y or kind==ROTATION_Z or
          kind==ROTATION_U1 or kind==ROTATION_U2 or kind==ROTATION_U3):
        return 1
    elif (kind==CONTROLLED_X or kind==CONTROLLED_Y or kind==CONTROLLED_Z or
          kind==CONTROLLED_XR or kind==CONTROLLED_XR_ or kind==CONTROLLED_H or
          kind==CONTROLLED_S or kind==CONTROLLED_S_ or kind==CONTROLLED_T or
          kind==CONTROLLED_T_ or kind==SWAP_QUBITS or kind==CONTROLLED_P or
          kind==CONTROLLED_RX or kind==CONTROLLED_RY or kind==CONTROLLED_RZ or
          kind==CONTROLLED_U1 or kind==CONTROLLED_U2 or kind==CONTROLLED_U3 or
          kind==MEASURE_BELL):
        return 2
    else:
        raise QState_UnknownQgateKind()

def get_qgate_param_num(kind=None):

    if (kind==SHOW or kind==BLOCH or
        kind==MEASURE or kind==MEASURE_X or kind==MEASURE_Y or kind==MEASURE_Z or kind==MEASURE_BELL or
        kind==RESET or
        kind==PAULI_X or kind==PAULI_Y or kind==PAULI_Z or
        kind==ROOT_PAULI_X or kind==ROOT_PAULI_X_ or kind==HADAMARD or
        kind==PHASE_SHIFT_S or kind==PHASE_SHIFT_S_ or kind==PHASE_SHIFT_T or kind==PHASE_SHIFT_T_ or
        kind==CONTROLLED_X or kind==CONTROLLED_Y or kind==CONTROLLED_Z or
        kind==CONTROLLED_XR or kind==CONTROLLED_XR_ or kind==CONTROLLED_H or
        kind==CONTROLLED_S or kind==CONTROLLED_S_ or kind==CONTROLLED_T or kind==CONTROLLED_T_ or kind==SWAP_QUBITS):
        return 0
    elif (kind==PHASE_SHIFT or
          kind==ROTATION_X or kind==ROTATION_Y or kind==ROTATION_Z or
          kind==ROTATION_U1 or kind==CONTROLLED_P or
          kind==CONTROLLED_RX or kind==CONTROLLED_RY or kind==CONTROLLED_RZ or
          kind==CONTROLLED_U1):
        return 1
    elif (kind==ROTATION_U2 or kind==CONTROLLED_U2):
        return 2
    elif (kind==ROTATION_U3 or kind==CONTROLLED_U3):
        return 3
    else:
        raise QState_UnknownQgateKind()

def is_clifford_gate(kind):

    if kind in (PAULI_X, PAULI_Y, PAULI_Z, HADAMARD,
                PHASE_SHIFT_S, PHASE_SHIFT_S_,
                CONTROLLED_X, CONTROLLED_Y, CONTROLLED_Z):
        return True
    else:
        return False

def is_non_clifford_gate(kind):

    if is_clifford_gate(kind) == False and is_measurement_gate(kind) == False and is_reset_gate(kind) == False:
        return True
    else:
        return False

def is_measurement_gate(kind):

    if kind in (MEASURE, MEASURE_X, MEASURE_Y, MEASURE_Z, MEASURE_BELL):
        return True
    else:
        return False

def is_reset_gate(kind):

    if kind == RESET:
        return True
    else:
        return False

def reverse_bit_order(vec_in):

    vec_out = [0] * len(vec_in)
    digits = np.log2(len(vec_in))
    if not digits.is_integer():
        raise ValueError("length of vector must be power of two")
        
    for idx, val in enumerate(vec_in):
        idx_binstr = '{:0{digits}b}'.format(idx, digits=int(digits))
        idx_binstr_rev = ''.join(reversed(list(idx_binstr)))
        idx_rev = int(idx_binstr_rev, 2)
        vec_out[idx_rev] = val

    return vec_out



# -*- coding: utf-8 -*-
from qlazy.config import *
from qlazy.error import *
from qlazy.util import *

# multi-controlled X gate

def __gray_code(qs, n):

    for k in range(2**n):
        yield k^(k>>1)

def qstate_mcx(qs,qid=[]):

    # controled and target register
    qid_ctr = qid[:-1]
    qid_tar = qid[-1]
        
    # hadamard
    qs.h(qid_tar)

    # controlled-RZ(psi), psi=pi/(2**(bitnum-1))
    bitnum = len(qid_ctr)
    psi = 1.0/(2**(bitnum-1)) # unit=pi(radian)
    gray_pre = 0
    for gray in __gray_code(qs,bitnum):
        if gray == 0:
            continue
        msb = len(str(bin(gray)))-3
        chb = len(str(bin(gray^gray_pre)))-3
        if gray != 1:
            if chb == msb:
                chb -= 1
            qs.cx(qid_ctr[chb],qid_ctr[msb])
        qs.cp(qid_ctr[msb],qid_tar,phase=psi)
        psi = -psi
        gray_pre = gray
    
    # hadamard
    qs.h(qid_tar)



# -*- coding: utf-8 -*-
import ctypes
from ctypes.util import find_library
import numpy as np
import pathlib
from collections import Counter 

from qlazy.config import *
from qlazy.error import *
from qlazy.util import *
from qlazy.QState import QState
from qlazy.MData import *
from qlazy.Observable import Observable
from qlazy.QCirc import QCirc
from qlazy.CMem import CMem
from qlazy.lib.mdata_c import *

lib= ctypes.CDLL(str(pathlib.Path(__file__).with_name('libqlz.'+get_lib_ext())))
libc = ctypes.CDLL(find_library("c"),mode=ctypes.RTLD_GLOBAL)

def qstate_init(qubit_num=None, seed=None):

    libc.srand(ctypes.c_int(seed))
        
    qstate = None
    c_qstate = ctypes.c_void_p(qstate)

    lib.qstate_init.restype = ctypes.c_int
    lib.qstate_init.argtypes = [ctypes.c_int, ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.qstate_init(ctypes.c_int(qubit_num), c_qstate)

    if ret == FALSE:
        raise QState_Error_Initialize()

    return c_qstate

def qstate_init_with_vector(vector=None, seed=None):
        
    libc.srand(ctypes.c_int(seed))
        
    qstate = None
    c_qstate = ctypes.c_void_p(qstate)
    
    dim = len(vector)
    vec_real = [0.0 for _ in range(dim)]
    vec_imag = [0.0 for _ in range(dim)]
    for i in range(dim):
        vec_real[i] = vector[i].real
        vec_imag[i] = vector[i].imag
    
    DoubleArray = ctypes.c_double * dim
    c_vec_real = DoubleArray(*vec_real)
    c_vec_imag = DoubleArray(*vec_imag)
    
    lib.qstate_init_with_vector.restype = ctypes.c_int
    lib.qstate_init_with_vector.argtypes = [DoubleArray, DoubleArray, ctypes.c_int,
                                            ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.qstate_init_with_vector(c_vec_real, c_vec_imag, ctypes.c_int(dim),
                                      c_qstate)
    
    if ret == FALSE:
        raise QState_Error_Initialize()
    
    return c_qstate


def qstate_reset(qs, qid=None):
    
    if qid is None or qid == []:
        qid = [i for i in range(qs.qubit_num)]
    
    qstate_check_args(qs, kind=SHOW, shots=None, angle=None, qid=qid)
    
    try:
        qubit_num = len(qid)
        qubit_id = [0 for _ in range(qubit_num)]
        for i in range(len(qid)):
            qubit_id[i] = qid[i]
    
        IntArray = ctypes.c_int * qubit_num
        qid_array = IntArray(*qubit_id)
            
        lib.qstate_reset.restype = ctypes.c_int
        lib.qstate_reset.argtypes = [ctypes.POINTER(QState),ctypes.c_int, IntArray]
        ret = lib.qstate_reset(ctypes.byref(qs),ctypes.c_int(qubit_num), qid_array)
    
        if ret == FALSE:
            raise QState_Error_Reset()
    
    except Exception:
        raise QState_Error_Reset()

    
def qstate_print(qs, qid=None, nonzero=False):

    if qid is None or qid == []:
        qid = [i for i in range(qs.qubit_num)]

    qstate_check_args(qs, kind=SHOW, shots=None, angle=None, qid=qid)
        
    try:
        qubit_num = len(qid)
        qubit_id = [0 for _ in range(qubit_num)]
        for i in range(len(qid)):
            qubit_id[i] = qid[i]

        IntArray = ctypes.c_int * qubit_num
        qid_array = IntArray(*qubit_id)

        
            
        lib.qstate_print.restype = ctypes.c_int
        lib.qstate_print.argtypes = [ctypes.POINTER(QState),ctypes.c_int, IntArray, ctypes.c_bool]
        ret = lib.qstate_print(ctypes.byref(qs),ctypes.c_int(qubit_num), qid_array, ctypes.c_bool(nonzero))

        if ret == FALSE:
            raise QState_Error_Show()

    except Exception:
        raise QState_Error_Show()


def qstate_copy(qs):

    try:
        qstate = None
        c_qstate = ctypes.c_void_p(qstate)
            
        lib.qstate_copy.restype = ctypes.c_int
        lib.qstate_copy.argtypes = [ctypes.POINTER(QState),
                                    ctypes.POINTER(ctypes.c_void_p)]
        ret = lib.qstate_copy(ctypes.byref(qs), c_qstate)

        if ret == FALSE:
            raise QState_Error_Clone()

        return c_qstate
        
    except Exception:
        raise QState_Error_Clone()


def qstate_bloch(qs, q=0):

    # error check
    qstate_check_args(qs, kind=BLOCH, shots=None, angle=None, qid=[q])

    try:
        theta = 0.0
        phi = 0.0
        c_theta = ctypes.c_double(theta)
        c_phi = ctypes.c_double(phi)
            
        lib.qstate_bloch.restype = ctypes.c_int
        lib.qstate_bloch.argtypes = [ctypes.POINTER(QState),ctypes.c_int,
                                     ctypes.POINTER(ctypes.c_double),
                                     ctypes.POINTER(ctypes.c_double)]
        ret = lib.qstate_bloch(ctypes.byref(qs),ctypes.c_int(q),
                               ctypes.byref(c_theta), ctypes.byref(c_phi))

        if ret == FALSE:
            raise QState_Error_Bloch()

        theta = c_theta.value
        phi = c_phi.value

        return theta,phi

    except Exception:
        raise QState_Error_Bloch()


def qstate_inner_product(qs_0, qs_1):

    try:
        real = 0.0
        imag = 0.0
        c_real = ctypes.c_double(real)
        c_imag = ctypes.c_double(imag)
            
        lib.qstate_inner_product.restype = ctypes.c_int
        lib.qstate_inner_product.argtypes = [ctypes.POINTER(QState),
                                             ctypes.POINTER(QState),
                                             ctypes.POINTER(ctypes.c_double),
                                             ctypes.POINTER(ctypes.c_double)]
        ret = lib.qstate_inner_product(ctypes.byref(qs_0),ctypes.byref(qs_1),
                                       ctypes.byref(c_real), ctypes.byref(c_imag))

        if ret == FALSE:
            raise QState_Error_InnerProduct()

        real = c_real.value
        imag = c_imag.value

        return complex(real, imag)
        
    except Exception:
        raise QState_Error_InnerProduct()

    
def qstate_get_camp(qs, qid=None):

    if qid is None or qid == []:
        qid = [i for i in range(qs.qubit_num)]

    # error check
    if len(qid) > qs.qubit_num:
        raise QState_TooManyArguments()
    for i in range(len(qid)):
        if qid[i] >= qs.qubit_num:
            raise QState_OutOfBound()
        if qid[i] < 0:
            raise QState_OutOfBound()

    try:
        qubit_num = len(qid)
        qubit_id = [0 for _ in range(qubit_num)]
        for i in range(len(qid)):
            qubit_id[i] = qid[i]
        IntArray = ctypes.c_int * qubit_num
        qid_array = IntArray(*qubit_id)

        camp = None
        c_camp = ctypes.c_void_p(camp)
        lib.qstate_get_camp.restype = ctypes.c_int
        lib.qstate_get_camp.argtypes = [ctypes.POINTER(QState),ctypes.c_int, IntArray,
                                        ctypes.POINTER(ctypes.c_void_p)]
        ret = lib.qstate_get_camp(ctypes.byref(qs),ctypes.c_int(qubit_num),
                                  qid_array, c_camp)

        if ret == FALSE:
            raise QState_Error_GetAmp()
                
        o = ctypes.cast(c_camp.value, ctypes.POINTER(ctypes.c_double))
            
        state_num = (1 << len(qid))
        out = [0] * state_num
        for i in range(state_num):
            out[i] = complex(o[2*i],o[2*i+1])

        libc.free.argtypes = [ctypes.POINTER(ctypes.c_double)]
        libc.free(o)

    except Exception:
        raise QState_Error_GetCmp()

    return np.array(out)


def qstate_tensor_product(qs, qstate):

    try:
        qstate_out = None
        c_qstate_out = ctypes.c_void_p(qstate_out)

        lib.qstate_tensor_product.restype = ctypes.c_int
        lib.qstate_tensor_product.argtypes = [ctypes.POINTER(QState),
                                              ctypes.POINTER(QState),
                                              ctypes.POINTER(ctypes.c_void_p)]
        ret = lib.qstate_tensor_product(ctypes.byref(qs),ctypes.byref(qstate),
                                        c_qstate_out)

        if ret == FALSE:
            raise QState_Error_TensorProduct()

        return c_qstate_out

    except Exception:
        raise QState_Error_TensorProduct()


def qstate_evolve(qs, observable=None, time=0.0, iter=0):

    if iter < 1:
        raise QState_Error_Evolve()
        
    if observable is None:
        raise QState_Error_Evolve()
        
    try:
        lib.qstate_evolve.restype = ctypes.c_int
        lib.qstate_evolve.argtypes = [ctypes.POINTER(QState),ctypes.POINTER(Observable),
                                      ctypes.c_double, ctypes.c_int]
        ret = lib.qstate_evolve(ctypes.byref(qs), ctypes.byref(observable),
                                ctypes.c_double(time), ctypes.c_int(iter))

        if ret == FALSE:
            raise QState_Error_Evolve()
            
    except Exception:
        raise QState_Error_Evolve()


def qstate_expect_value(qs, observable=None):

    if observable is None:
        raise QState_Error_Expect()
        
    try:
        val = 0.0
        c_val = ctypes.c_double(val)
        lib.qstate_expect_value.restype = ctypes.c_int
        lib.qstate_expect_value.argtypes = [ctypes.POINTER(QState),
                                            ctypes.POINTER(Observable),
                                            ctypes.POINTER(ctypes.c_double)]
        ret = lib.qstate_expect_value(ctypes.byref(qs),
                                      ctypes.byref(observable),
                                      ctypes.byref(c_val))
            
        if ret == FALSE:
            raise QState_Error_Expect()

        val = c_val.value
            
        return complex(val,0.0)
            
    except Exception:
        raise QState_Error_Expect()

    return out


def qstate_apply_matrix(qs, matrix=None, qid=None):

    if matrix is None:
        raise QState_Error_Apply()
        
    if qid is None or qid == []:
        qid = [i for i in range(qs.qubit_num)]

    try:
        qubit_num = len(qid)
        qubit_id = [0 for _ in range(qubit_num)]
        for i in range(len(qid)):
            qubit_id[i] = qid[i]
        IntArray = ctypes.c_int * qubit_num
        qid_array = IntArray(*qubit_id)

        row = len(matrix) # dimension of the unitary matrix
        col = row
        size = row * col

        mat_complex = []
        for mat_row in matrix:
            for mat_elm in mat_row:
                mat_complex.append(mat_elm)
        
        mat_real = [0.0 for _ in range(size)]
        mat_imag = [0.0 for _ in range(size)]
        for i in range(size):
            mat_real[i] = mat_complex[i].real
            mat_imag[i] = mat_complex[i].imag
                
        DoubleArray = ctypes.c_double * size
        c_mat_real = DoubleArray(*mat_real)
        c_mat_imag = DoubleArray(*mat_imag)
            
        lib.qstate_apply_matrix.restype = ctypes.c_int
        lib.qstate_apply_matrix.argtypes = [ctypes.POINTER(QState),
                                            ctypes.c_int, IntArray,
                                            DoubleArray, DoubleArray,
                                            ctypes.c_int, ctypes.c_int]
        ret = lib.qstate_apply_matrix(ctypes.byref(qs),
                                      ctypes.c_int(qubit_num), qid_array,
                                      c_mat_real, c_mat_imag,
                                      ctypes.c_int(row), ctypes.c_int(col))

        if ret == FALSE:
            raise QState_Error_Apply()

    except Exception:
        raise QState_Error_Apply()

    
def qstate_operate_qgate(qs, kind=None, qid=None,
                         phase=DEF_PHASE, phase1=DEF_PHASE, phase2=DEF_PHASE):

    # error check
    qstate_check_args(qs, kind=kind, qid=qid, shots=None, angle=None,
                      phase=phase, phase1=phase1, phase2=phase2)

    qubit_id = [-1 for _ in range(2)]
    for i in range(len(qid)):
        qubit_id[i] = qid[i]
    IntArray = ctypes.c_int * 2
    qid_array = IntArray(*qubit_id)

    lib.qstate_operate_qgate.restype = ctypes.c_int
    lib.qstate_operate_qgate.argtypes = [ctypes.POINTER(QState), ctypes.c_int,
                                         ctypes.c_double, ctypes.c_double,
                                         ctypes.c_double, IntArray]
    ret = lib.qstate_operate_qgate(ctypes.byref(qs), ctypes.c_int(kind),
                                   ctypes.c_double(phase), ctypes.c_double(phase1),
                                   ctypes.c_double(phase2), qid_array)

    if ret == FALSE:
        raise QState_Error_OperateQgate()

def qstate_measure(qs, qid=None, angle=0.0, phase=0.0):

    if qid is None or qid == []:
        qid = [i for i in range(qs.qubit_num)]

    # operate
    qubit_num = len(qid)
    qubit_id = [0 for _ in range(qubit_num)]
    for i in range(len(qid)):
        qubit_id[i] = qid[i]
    IntArray = ctypes.c_int * qubit_num
    qid_array = IntArray(*qubit_id)

    mval = 0
    c_mval = ctypes.c_int(mval)

    lib.qstate_measure.restype = ctypes.c_int
    lib.qstate_measure.argtypes = [ctypes.POINTER(QState),
                                   ctypes.c_double, ctypes.c_double,
                                   ctypes.c_int, IntArray, ctypes.POINTER(ctypes.c_int)]
    ret = lib.qstate_measure(ctypes.byref(qs),
                             ctypes.c_double(angle), ctypes.c_double(phase),
                             ctypes.c_int(qubit_num), qid_array, ctypes.byref(c_mval))

    if ret == FALSE:
        raise QState_Error_Measure()

    mval = c_mval.value
    digits = len(qid)
    mval = '{:0{digits}b}'.format(mval, digits=digits)

    return mval

def qstate_measure_stats(qs, qid=None, shots=DEF_SHOTS, angle=0.0, phase=0.0):

    if qid is None or qid == []:
        qid = [i for i in range(qs.qubit_num)]

    # error check
    qstate_check_args(qs, kind=MEASURE, qid=qid, shots=shots, angle=angle, phase=phase)

    # operate
    qubit_num = len(qid)
    qubit_id = [0 for _ in range(qubit_num)]
    for i in range(len(qid)):
        qubit_id[i] = qid[i]
    IntArray = ctypes.c_int * qubit_num
    qid_array = IntArray(*qubit_id)

    mdata = None
    c_mdata = ctypes.c_void_p(mdata)

    lib.qstate_measure_stats.restype = ctypes.c_int
    lib.qstate_measure_stats.argtypes = [ctypes.POINTER(QState), ctypes.c_int,
                                         ctypes.c_double, ctypes.c_double,
                                         ctypes.c_int, IntArray,
                                         ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.qstate_measure_stats(ctypes.byref(qs), ctypes.c_int(shots),
                                   ctypes.c_double(angle), ctypes.c_double(phase),
                                   ctypes.c_int(qubit_num), qid_array, c_mdata)

    if ret == FALSE:
        raise QState_Error_Measure()

    out = ctypes.cast(c_mdata.value, ctypes.POINTER(MData))
        
    return out.contents

def qstate_measure_bell_stats(qs, qid=None, shots=DEF_SHOTS):

    if qid is None or qid == []:
        qid = [i for i in range(2)]
            
    # error check
    qstate_check_args(qs, kind=MEASURE_BELL, qid=qid, shots=shots, angle=None, phase=None)

    # operate
    state_num = 4
    qubit_num = 2
    qubit_id = [0 for _ in range(qubit_num)]
    for i in range(qubit_num):
        qubit_id[i] = qid[i]
    IntArray = ctypes.c_int * qubit_num
    qid_array = IntArray(*qubit_id)
        
    mdata = None
    c_mdata = ctypes.c_void_p(mdata)

    lib.qstate_measure_bell_stats.restype = ctypes.c_int
    lib.qstate_measure_bell_stats.argtypes = [ctypes.POINTER(QState), ctypes.c_int,
                                              ctypes.c_int, IntArray,
                                              ctypes.POINTER(ctypes.c_void_p)]

    ret = lib.qstate_measure_bell_stats(ctypes.byref(qs), ctypes.c_int(shots),
                                        ctypes.c_int(qubit_num), qid_array, c_mdata)

    if ret == FALSE:
        raise QState_Error_Measure()

    out = ctypes.cast(c_mdata.value, ctypes.POINTER(MData))
        
    return out.contents

def qstate_operate_qcirc(qstate, cmem, qcirc, shots, cid):

    lib.qstate_operate_qcirc.restype = ctypes.c_int
    lib.qstate_operate_qcirc.argtypes = [ctypes.POINTER(QState), ctypes.POINTER(CMem), ctypes.POINTER(QCirc)]

    if cmem is not None:

        cmem_num = cmem.cmem_num
        frequency = Counter()
        for n in range(shots):
            
            if n < shots - 1:
                qstate_tmp = qstate.clone()
                cmem_tmp = cmem.clone()
                ret = lib.qstate_operate_qcirc(ctypes.byref(qstate_tmp), ctypes.byref(cmem_tmp), ctypes.byref(qcirc))
                bit_array = ctypes.cast(cmem_tmp.bit_array, ctypes.POINTER(ctypes.c_int*cmem_num))
            else:
                ret = lib.qstate_operate_qcirc(ctypes.byref(qstate), ctypes.byref(cmem), ctypes.byref(qcirc))
                bit_array = ctypes.cast(cmem.bit_array, ctypes.POINTER(ctypes.c_int*cmem_num))

            if ret == FALSE:
                raise QState_Error_OperateQcirc()
        
            cmem_list = [bit_array.contents[i] for i in range(cmem_num)]
            cmem_list_part = [cmem_list[c] for c in cid]
            mval = "".join(map(str, cmem_list_part))

            frequency[mval] += 1

        return frequency

    else:

        c_cmem = ctypes.POINTER(CMem)()
        ret = lib.qstate_operate_qcirc(ctypes.byref(qstate), c_cmem, ctypes.byref(qcirc))

        if ret == FALSE:
            raise QState_Error_OperateQcirc()

        return None

def qstate_free(qs):

    lib.qstate_free.argtypes = [ctypes.POINTER(QState)]
    lib.qstate_free(ctypes.byref(qs))

def __gray_code(qs, n):

    for k in range(2**n):
        yield k^(k>>1)

def qstate_mcx(qs,qid=[]):

    # controled and target register
    qid_ctr = qid[:-1]
    qid_tar = qid[-1]
        
    # hadamard
    qs.h(qid_tar)

    # controlled-RZ(psi), psi=pi/(2**(bitnum-1))
    bitnum = len(qid_ctr)
    psi = 1.0/(2**(bitnum-1)) # unit=pi(radian)
    gray_pre = 0
    for gray in __gray_code(qs,bitnum):
        if gray == 0:
            continue
        msb = len(str(bin(gray)))-3
        chb = len(str(bin(gray^gray_pre)))-3
        if gray != 1:
            if chb == msb:
                chb -= 1
            qs.cx(qid_ctr[chb],qid_ctr[msb])
        qs.cp(qid_ctr[msb],qid_tar,phase=psi)
        psi = -psi
        gray_pre = gray
    
    # hadamard
    qs.h(qid_tar)



# -*- coding: utf-8 -*-
import ctypes
from ctypes.util import find_library
import pathlib

from qlazy.config import *
from qlazy.error import *
from qlazy.util import *
from qlazy.MData import *

lib= ctypes.CDLL(str(pathlib.Path(__file__).with_name('libqlz.'+get_lib_ext())))
libc = ctypes.CDLL(find_library("c"),mode=ctypes.RTLD_GLOBAL)

def mdata_init(qubit_num=None, shots=1, angle=0.0, phase=0.0, qid=None):

    if qid is None or qid == []:
        qid = [i for i in range(qubit_num)]
        
    IntArray = ctypes.c_int * MAX_QUBIT_NUM
    qid_array = IntArray(*qubit_id)
    
    mdata = None
    c_mdata = ctypes.c_void_p(mdata)

    lib.mdata_init.restype = ctypes.c_int
    lib.mdata_init.argtypes = [ctypes.c_int, ctypes.c_int, ctypes.c_double, ctypes.c_double,
                               IntArray, ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.mdata_init(ctypes.c_int(qubit_num), ctypes.c_int(shots),
                         ctypes.c_double(angle), ctypes.c_double(phase), qid_array, c_mdata)

    if ret == FALSE:
        raise MData_Error_Initialize()

    return c_mdata


def mdata_print(md):

    lib.mdata_print.restype = ctypes.c_int
    lib.mdata_print.argtypes = [ctypes.POINTER(MData)]
    ret = lib.mdata_print(ctypes.byref(md))

    if ret == FALSE:
        raise MData_Error_Show()

def mdata_free(md):

    lib.mdata_free.argtypes = [ctypes.POINTER(MData)]
    lib.mdata_free(ctypes.byref(md))



# -*- coding: utf-8 -*-
import numpy as np

def CreateRegister(*args):
    """
    create multi-dimensional register id and initialize zero.

    Arguments
    ----------
    args : (int, int, ...)
        size of register id.

    Returns
    -------
    reg : list (multi-dimensional)
        register id list.

    Examples
    --------
    >>> reg = CreateRegister(2)
    >>> print(reg)
    [0,0]
    >>> reg = CreateRegister(2,3)
    >>> print(reg)
    [[0,0,0],[0,0,0]]
        
    """
    global COUNT
    COUNT = 0
    reg = np.zeros(args, dtype=int).tolist()
    return reg

def InitRegister(*args):
    """
    initialize register id

    Arguments
    ---------
    args : list, list,...
        arguments of register id

    Returns
    -------
    total_num : int
       total number of registers

    Examples
    --------
    >>> reg_0 = QState.create_register(3)
    >>> reg_1 = QState.create_register(2,4)
    >>> print(reg_0, reg_1)
    [0,0,0] [[0,0,0,0],[0,0,0,0]]
    >>> reg_num = InitRegister(reg_0, reg_1)
    >>> print(reg_num, reg_0, reg_1)
    11 [0,1,2] [[3,4,5,6],[7,8,9,10]]
    """
    global COUNT
    for x in args:
        if type(x) is list:
            for i,v in enumerate(x):
                if type(v) is int:
                    x[i] = COUNT
                    COUNT += 1
                else:
                    num = InitRegister(v)
    return COUNT
